# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Fast access to the auto optimization on DaCe."""

from typing import Any, Optional, Sequence

import dace
from dace import data as dace_data
from dace.sdfg import nodes as dace_nodes, propagation as dace_propagation
from dace.transformation.auto import auto_optimize as dace_aoptimize
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next import common as gtx_common
from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


def gt_auto_optimize(
    sdfg: dace.SDFG,
    gpu: bool,
    unit_strides_kind: Optional[gtx_common.DimensionKind] = None,
    make_persistent: bool = False,
    gpu_block_size: Optional[Sequence[int | str] | str] = None,
    blocking_dim: Optional[gtx_common.Dimension] = None,
    blocking_size: int = 10,
    blocking_only_if_independent_nodes: bool = True,
    reuse_transients: bool = False,
    gpu_launch_bounds: Optional[int | str] = None,
    gpu_launch_factor: Optional[int] = None,
    gpu_memory_pool: bool = False,
    constant_symbols: Optional[dict[str, Any]] = None,
    assume_pointwise: bool = True,
    validate: bool = True,
    validate_all: bool = False,
    **kwargs: Any,
) -> dace.SDFG:
    """Performs GT4Py specific optimizations on the SDFG in place.

    This function expects that the input SDFG follows the principles that are
    outlined in [ADR0018](https://github.com/GridTools/gt4py/tree/main/docs/development/ADRs/0018-Canonical_SDFG_in_GT4Py_Transformations.md).
    It is important to notice, that if `reuse_transients` is active then the
    optimized SDFG no longer conforms to these rules.

    The auto optimization works in different phases, that focuses each on
    different aspects of the SDFG. The initial SDFG is assumed to have a
    very large number of rather simple Maps.

    Note, because of how `gt_auto_optimizer()` works it is not save to call
    it twice on the same SDFG.

    A user should not rely on a particular order in which the SDFG is processed.
    However, in broad terms the function performs roughly the following steps:
    - An initial simplification is performed. This also includes the substitution
        of symbols whose value is known, see `constant_symbols`.
    - The top level Maps are optimized, the goal of this step is to increase
        the computational intensity. This is mostly, but not limited to fusing
        then together.
    - The interior dataflow of the top level Maps are optimized.
    - The Maps are configured. This involves setting their iteration order, such
        that the unit stride dimension is associated to the `x` block (GPU) or
        inner most loop-nest (CPU). But it also involves moving everything to GPU,
        if needed and setting the block size.
    - The transients are processed, this involves steps such as determine their
        strides or if they are persistently allocated.

    Args:
        sdfg: The SDFG that should be optimized in place.
        gpu: Optimize for GPU or CPU.
        unit_strides_kind: All dimensions of this kind are considered to have unit
            strides, see `gt_set_iteration_order()` for more.
        make_persistent: Turn all transients to persistent lifetime, thus they are
            allocated over the whole lifetime of the program, even if the kernel exits.
            Thus the SDFG can not be called by different threads.
        gpu_block_size: The thread block size for maps in GPU mode, currently only
            one for all.
        blocking_dim: On which dimension blocking should be applied.
        blocking_size: How many elements each block should process.
        blocking_only_if_independent_nodes: If `True`, the default, only apply loop
            blocking if there are independent nodes in the Map, see the
            `require_independent_nodes` option of the `LoopBlocking` transformation.
        reuse_transients: Run the `TransientReuse` transformation, might reduce memory footprint.
        gpu_launch_bounds: Use this value as `__launch_bounds__` for _all_ GPU Maps.
        gpu_launch_factor: Use the number of threads times this value as `__launch_bounds__`
            for _all_ GPU Maps.
        gpu_memory_pool: Enable CUDA memory pool in gpu codegen.
        constant_symbols: Symbols listed in this `dict` will be replaced by the
            respective value inside the SDFG. This might increase performance.
        assume_pointwise: Assume that the SDFG has no risk for race condition in
            global data access. See the `GT4PyMapBufferElimination` transformation for more.
        validate: Perform validation during the steps.
        validate_all: Perform extensive validation.

    Note:
        - For identifying symbols that can be treated as compile time constants
            `gt_find_constant_arguments()` function can be used.
        - When `unit_strides_kind` is `None` the function assumes that on GPU the
            horizontal dimension has unit strides, while on CPU the vertical dimension
            has unit strides.
        - In case GPU optimizations are enabled, the function assumes that all
            global fields are already on the GPU.

    Todo:
        - Update the description. The Phases are nice, but they have lost their
            link to reality a little bit.
        - Improve the determination of the strides and iteration order of the
            transients.
        - Set padding of transients, i.e. alignment, the DaCe datadescriptor
            can do that.
        - Handle nested SDFGs better.
        - Specify arguments to set the size of GPU thread blocks depending on the
            dimensions. I.e. be able to use a different size for 1D than 2D Maps.
        - Implement some model to further guide to determine what we want to fuse.
            Something along the line "Fuse if operational intensity goes up, but
            not if we have too much internal space (register pressure).
    """
    device = dace.DeviceType.GPU if gpu else dace.DeviceType.CPU

    with dace.config.temporary_config():
        # Do not store which transformations were applied inside the SDFG.
        dace.Config.set("store_history", value=False)

        # If there is an exception during `can_be_applied()` propagate it instead
        #  of ignoring it.
        dace.Config.set("optimizer", "match_exception", value=True)

        # Do not assume that symbols are positive. The only reason for this is
        #  that the origin can potentially be negative.
        #  See also [issue#2095](https://github.com/spcl/dace/issues/2095)
        dace.Config.set("optimizer", "symbolic_positive", value=False)

        # Initial Cleanup
        # NOTE: The initial simplification stage must be synchronized with the one that
        #   `gt_substitute_compiletime_symbols()` performs!
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=False,
            skip=gtx_transformations.constants._GT_AUTO_OPT_INITIAL_STEP_SIMPLIFY_SKIP_LIST,
            validate_all=validate_all,
        )

        if constant_symbols:
            gtx_transformations.gt_substitute_compiletime_symbols(
                sdfg=sdfg,
                repl=constant_symbols,
                simplify=True,  # Simplify again after.
                skip=gtx_transformations.constants._GT_AUTO_OPT_INITIAL_STEP_SIMPLIFY_SKIP_LIST,
                simplify_at_entry=False,
                validate=False,
                validate_all=validate_all,
            )

        gtx_transformations.gt_reduce_distributed_buffering(
            sdfg, validate=False, validate_all=validate_all
        )

        # Process top level Maps
        sdfg = _gt_auto_process_top_level_maps(
            sdfg=sdfg,
            assume_pointwise=assume_pointwise,
            validate_all=validate_all,
        )

        # We now ensure that point wise computations are properly double buffered,
        #  this ensures that rule 3 of ADR-18 is maintained.
        # TODO(phimuell): Figuring out if it is important to do it before the inner
        #   Map optimization. I think it is, especially when we apply `LoopBlocking`.
        gtx_transformations.gt_create_local_double_buffering(sdfg)

        # Optimize the interior of the Maps:
        sdfg = _gt_auto_process_dataflow_inside_maps(
            sdfg=sdfg,
            blocking_dim=blocking_dim,
            blocking_size=blocking_size,
            blocking_only_if_independent_nodes=blocking_only_if_independent_nodes,
            validate_all=validate_all,
        )

        # Configure the Maps:
        #  Will also perform the GPU transformation.
        # TODO(phimuell): Maybe switch it with the inside map optimization.
        sdfg = _gt_auto_configure_maps_and_strides(
            sdfg=sdfg,
            gpu=gpu,
            unit_strides_kind=unit_strides_kind,
            gpu_block_size=gpu_block_size,
            gpu_launch_factor=gpu_launch_factor,
            gpu_launch_bounds=gpu_launch_bounds,
            validate_all=validate_all,
        )

        # Transients
        sdfg = _gt_auto_post_processing(
            sdfg=sdfg,
            gpu=gpu,
            make_persistent=make_persistent,
            # TODO(phimuell): In general `TransientReuse` is a good idea, but the
            #   current implementation also reuses transients scalars inside Map
            #   scopes, which I do not like. Thus we should fix the transformation
            #   to avoid that.
            reuse_transients=reuse_transients,
            gpu_memory_pool=gpu_memory_pool,
            validate_all=validate_all,
        )

        # Set the implementation of the library nodes.
        dace_aoptimize.set_fast_implementations(sdfg, device)

        if validate:
            sdfg.validate()

        return sdfg


def _gt_auto_process_top_level_maps(
    sdfg: dace.SDFG,
    assume_pointwise: bool,
    validate_all: bool,
) -> dace.SDFG:
    """Optimize the Maps at the top level of the SDFG inplace.

    In broad terms the function tries to reduce the number of of kernels/maps
    by fusing/inlining them, however, the function will also perform some other
    transformations, with the goal to promote high computational intensity and
    occupancy (on GPU).

    It is important to note that the fusion will only affect top level maps, i.e.
    everything whose scope is `None`. Some parts of the optimization process are
    iteratively.

    For a full description of the argument see the `gt_auto_optimize()` function.

    Note:
        The function assumes that `gt_simplify()` has been called on the SDFG
        before it is passed to this function.
    """

    # NOTE: Inside this function we have to disable the consolidation of edges.
    #   This is because it might block the application of `SpliAccessNode`. As
    #   an example consider a Map that writes `tmp[:, 80]` and a second Map that
    #   writes `tmp[:, 0]`, if these two Maps are now horizontally fussed and
    #   edge consolidation is on, then the resulting Map would "write", at least
    #   according to the subset, after Memlet propagation, into `tmp[:, 0:80]`.
    #   For that reason we block edge consolidation inside this function.
    #   However, we allow allow the consolidation, in MapFusion if this does
    #   not lead to an extension. This is because it causes some issues with
    #   MapFusion.
    # TODO(phimuell): Find a better way as blocking edge consolidation might
    #   limit what MapFusion can do.
    # TODO(phimuell): Maybe disable edge consolidation by default?

    # Compute the SDFG hash to see if something has changed.
    #  We use the hash instead of the return values of the transformation, because
    #  computing the hash invalidates some caches that are not properly updated in Dace.
    # TODO(phimuell): Remove this hack as soon as DaCe is fixed.
    sdfg_hash = sdfg.hash_sdfg()

    while True:
        # First we do scan the entire SDFG to figure out which data is only
        #  used once and can be deleted.
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

        # TODO(phimuell): Use a cost measurement to decide if fusion should be done.

        vertical_map_fusion = gtx_transformations.MapFusionVertical(
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=True,
        )
        # TODO(phimuell): Remove that hack once [issue#1911](https://github.com/spcl/dace/issues/1911)
        #   has been solved.
        vertical_map_fusion._single_use_data = single_use_data

        horizontal_map_fusion = gtx_transformations.MapFusionHorizontal(
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=True,
            # TODO(phimuell): Should we enable this flag?
            only_if_common_ancestor=False,
        )

        # NOTE: Since horizontal Map fusion matches _any_ two Maps, running it on large
        #  SDFGs is expensive. Thus we run vertical Map fusion first to reduce the search
        #  space.
        sdfg.apply_transformations_repeated(
            vertical_map_fusion,
            validate=False,
            validate_all=validate_all,
        )
        sdfg.apply_transformations_repeated(
            [horizontal_map_fusion, vertical_map_fusion],
            validate=False,
            validate_all=validate_all,
        )

        # We have to call it here, such that some other transformations, most
        #  importantly the split transformations run.
        #  We allow promotion of vertical (which is beneficial for Nabla4 type
        #  stencils and if `LoopBlocking` is enabled. However, we also allow
        #  horizontal promotion, this is an empirical choice. If the computation
        #  on the horizontal are very complicated, it might degrade performance
        #  but such cases should be rare.
        # TODO(phimuell): Add a criteria to decide if we should promote or not.
        # TODO(phimuell): Think about relocating this before horizontal Map fusion.
        sdfg.apply_transformations_repeated(
            gtx_transformations.MapPromoter(
                only_toplevel_maps=True,
                promote_vertical=True,
                promote_horizontal=True,
                promote_local=False,
                promotion_callback=_check_if_horizontal_map_promotion_is_appropriate,
                single_use_data=single_use_data,
            ),
            validate=False,
            validate_all=validate_all,
        )

        # Now do some cleanup task, that may enable further fusion opportunities.
        #  Note for performance reasons simplify is deferred.
        gtx_transformations.gt_reduce_distributed_buffering(
            sdfg, validate=False, validate_all=validate_all
        )

        # TODO(phimuell): Find out how to skip the propagation and integrating it
        #   into the split transformation.
        sdfg.apply_transformations_repeated(
            gtx_transformations.SplitConsumerMemlet(single_use_data=single_use_data),
            validate=False,
            validate_all=validate_all,
        )
        dace_propagation.propagate_memlets_sdfg(sdfg)

        sdfg.apply_transformations_repeated(
            [
                # TODO(phimuell): The transformation is also active inside Maps.
                #   Which is against the description of this function, but it should
                #   not matter that much.
                gtx_transformations.SplitAccessNode(
                    single_use_data=single_use_data,
                ),
                gtx_transformations.GT4PyMapBufferElimination(
                    assume_pointwise=assume_pointwise,
                ),
            ],
            validate=False,
            validate_all=validate_all,
        )

        # Call vertical and horizontal map fusion to fuse together maps on partially
        #  overlapping range. This is an iterative process that splits the maps to
        #  expose overlapping range and applies serial/parallel map fusion.
        gtx_transformations.gt_vertical_map_split_fusion(
            sdfg=sdfg,
            run_simplify=False,
            skip=gtx_transformations.constants._GT_AUTO_OPT_TOP_LEVEL_STAGE_SIMPLIFY_SKIP_LIST,
            consolidate_edges_only_if_not_extending=True,
            single_use_data=single_use_data,
            validate=False,
            validate_all=validate_all,
        )
        gtx_transformations.gt_horizontal_map_split_fusion(
            sdfg=sdfg,
            run_simplify=False,
            skip=gtx_transformations.constants._GT_AUTO_OPT_TOP_LEVEL_STAGE_SIMPLIFY_SKIP_LIST,
            fuse_possible_maps=True,
            consolidate_edges_only_if_not_extending=True,
            validate=False,
            validate_all=validate_all,
        )

        # Determine if the SDFG has been modified by comparing the hash.
        old_sdfg_hash, sdfg_hash = sdfg_hash, sdfg.hash_sdfg()
        if old_sdfg_hash == sdfg_hash:
            break

        # The SDFG was modified by the transformations above. The SDFG was
        #  modified. Call Simplify and try again to further optimize.
        gtx_transformations.gt_simplify(
            sdfg,
            validate=False,
            validate_all=validate_all,
            skip=gtx_transformations.constants._GT_AUTO_OPT_TOP_LEVEL_STAGE_SIMPLIFY_SKIP_LIST,
        )

    return sdfg


def _gt_auto_process_dataflow_inside_maps(
    sdfg: dace.SDFG,
    blocking_dim: Optional[gtx_common.Dimension],
    blocking_size: int,
    blocking_only_if_independent_nodes: Optional[bool],
    validate_all: bool,
) -> dace.SDFG:
    """Optimizes the dataflow inside the top level Maps of the SDFG inplace.

    For a description of the arguments see `gt_auto_optimize()`.

    It is important that nested Maps are not fused together. The main reason is that
    in most cases it is not beneficial and might prevent other optimization, as the
    fusion (especially the parallel version) will create dependency of otherwise
    unrelated dataflow. Furthermore, the majority, if not all, of these maps are
    over a constant range, e.g. the number of neighbours, which is known at compile
    time, so the compiler will fully unroll them anyway.
    """

    # Constants (tasklets are needed to write them into a variable) should not be
    #  arguments to a kernel but be present inside the body.
    sdfg.apply_transformations_once_everywhere(
        gtx_transformations.GT4PyMoveTaskletIntoMap,
        validate=False,
        validate_all=validate_all,
    )
    gtx_transformations.gt_simplify(
        sdfg,
        skip=gtx_transformations.constants._GT_AUTO_OPT_INNER_DATAFLOW_STAGE_SIMPLIFY_SKIP_LIST,
        validate=False,
        validate_all=validate_all,
    )

    # Blocking is performed first, because this ensures that as much as possible
    #  is moved into the k independent part.
    if blocking_dim is not None:
        sdfg.apply_transformations_once_everywhere(
            gtx_transformations.LoopBlocking(
                blocking_size=blocking_size,
                blocking_parameter=blocking_dim,
                require_independent_nodes=blocking_only_if_independent_nodes,
            ),
            validate=False,
            validate_all=validate_all,
        )

    # Move dataflow into the branches of the `if` such that they are only evaluated
    #  if they are needed. Important to call it repeatedly.
    # TODO(phimuell): It is unclear if `MoveDataflowIntoIfBody` should be called
    #   before or after `LoopBlocking`. In cases where the condition is `False`
    #   most of the times calling it before is better, but if the condition is
    #   `True` then this order is better. Solve that issue.
    sdfg.apply_transformations_repeated(
        gtx_transformations.MoveDataflowIntoIfBody(
            ignore_upstream_blocks=False,
        ),
        validate=False,
        validate_all=validate_all,
    )
    gtx_transformations.gt_simplify(
        sdfg,
        skip=gtx_transformations.constants._GT_AUTO_OPT_INNER_DATAFLOW_STAGE_SIMPLIFY_SKIP_LIST,
        validate=False,
        validate_all=validate_all,
    )

    return sdfg


def _gt_auto_configure_maps_and_strides(
    sdfg: dace.SDFG,
    gpu: bool,
    unit_strides_kind: Optional[gtx_common.DimensionKind],
    gpu_block_size: Optional[Sequence[int | str] | str],
    gpu_launch_bounds: Optional[int | str],
    gpu_launch_factor: Optional[int],
    validate_all: bool,
) -> dace.SDFG:
    """Configure the Maps and the strides of the SDFG inplace.

    Essentially the function will set the properties of the Map accordingly, most
    importantly the iteration order of the loops, to ensure that the unit stride
    is associated to the `x` dimension of a block (GPU) or the inner most loop-nest.
    Furthermore, it will also apply the GPU transformation.
    It will also determine the strides of the transients.

    For a description of the arguments see the `gt_auto_optimize()` function.
    """

    # We now set the iteration order of the Maps. For that we use `unit_strides_kind`
    #  argument and if not supplied we guess depending if we are on the GPU or not.
    if unit_strides_kind is None:
        unit_strides_kind = (
            gtx_common.DimensionKind.HORIZONTAL if gpu else gtx_common.DimensionKind.VERTICAL
        )
    # It is not possible to use the `unit_strides_dim` argument of the
    #  function, because `LoopBlocking`, if run, changed the name of the
    #  parameter but the dimension can still be identified by its "kind".
    gtx_transformations.gt_set_iteration_order(
        sdfg=sdfg,
        unit_strides_kind=unit_strides_kind,
        validate=False,
        validate_all=validate_all,
    )

    # NOTE: We have to set the strides of transients before the non-standard Memlets
    #   get expanded, i.e. turned into Maps because no `cudaMemcpy*()` call exists,
    #   which requires that the final strides are there. Furthermore, Memlet expansion
    #   has to happen before the GPU block size is set. There are several possible
    #   solutions for that, of which none is really good. The one that is the least
    #   bad thing is to set the strides of the transients here. The main downside
    #   is that this and the `_gt_auto_post_processing()` function has these weird
    #   names.
    gtx_transformations.gt_change_transient_strides(sdfg, gpu=gpu)

    if gpu:
        # TODO(phimuell): The GPU function might modify the map iteration order.
        #   This is because how it is implemented (promotion and fusion). However,
        #   because of its current state, this should not happen, but we have to look
        #   into it.
        gtx_transformations.gt_gpu_transformation(
            sdfg,
            gpu_block_size=gpu_block_size,
            gpu_launch_bounds=gpu_launch_bounds,
            gpu_launch_factor=gpu_launch_factor,
            validate=False,
            validate_all=validate_all,
            try_removing_trivial_maps=True,
        )

    return sdfg


def _gt_auto_post_processing(
    sdfg: dace.SDFG,
    gpu: bool,
    make_persistent: bool,
    reuse_transients: bool,
    gpu_memory_pool: bool,
    validate_all: bool,
) -> dace.SDFG:
    """Perform post processing on the SDFG.

    Apply the finishing touch to the optimized SDFG.
    For a full description of the arguments see the `gt_auto_optimize()` function.
    """

    if reuse_transients:
        # TODO(phimuell): Investigate if we should enable it, it may make things
        #   harder for the compiler. Maybe write our own to only consider big
        #   transients and not small ones (~60B)
        transient_reuse = dace.transformation.passes.TransientReuse()
        transient_reuse.apply_pass(sdfg, {})

    # TODO(phimuell): Fix the bug, it uses the tile value and not the stack array value.
    dace_aoptimize.move_small_arrays_to_stack(sdfg)

    if make_persistent and gpu_memory_pool:
        raise ValueError("Cannot set both 'make_persistent' and 'gpu_memory_pool'.")

    if make_persistent:
        device = dace.DeviceType.GPU if gpu else dace.DeviceType.CPU
        gtx_transformations.gt_make_transients_persistent(sdfg=sdfg, device=device)

        if device == dace.DeviceType.GPU:
            # NOTE: For unknown reasons the counterpart of the
            #   `gt_make_transients_persistent()` function in DaCe, resets the
            #   `wcr_nonatomic` property of every memlet, i.e. makes it atomic.
            #   However, it does this only for edges on the top level and on GPU.
            #   For compatibility with DaCe (and until we found out why) the GT4Py
            #   auto optimizer will emulate this behaviour.
            for state in sdfg.states():
                assert isinstance(state, dace.SDFGState)
                for edge in state.edges():
                    edge.data.wcr_nonatomic = False

    if gpu and gpu_memory_pool:
        gtx_transformations.gpu_utils.gt_gpu_apply_mempool(sdfg)

    if validate_all:
        sdfg.validate()

    return sdfg


def _check_if_horizontal_map_promotion_is_appropriate(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    first_map_exit: dace_nodes.MapExit,
    second_map_entry: dace_nodes.MapEntry,
    missing_map_parameters: list[str],
) -> bool:
    """Helper function to check if horizontal Map promotion should be performed.

    The function is used by `_gt_auto_process_top_level_maps()` to ensure that
    only horizontal (the function will not check any other promotion) promotions
    that are beneficial will be accepted.

    The implementation checks if the complexity of the first map (the one that
    lacks the horizontal dimension) is not too large, such that inlining is
    beneficial. The rules, which are empirical, are summarized as follows:
    - The number of nodes inside the Map scope is below a certain value (currently
        20 nodes).
    - The Map only contain Tasklets and AccessNodes refering to transient data.
    - There are no WCR edges.

    For arguments see the `MapPromotionCallBack`.
    """

    # We only apply the rule to horizontal promotion. Vertical promotion is known
    #  to be beneficial, especially in conjunction with `LoopBloking`.
    if not any(
        missing_param.endswith("_gtx_horizontal") for missing_param in missing_map_parameters
    ):
        return True

    # NOTE: Inclusion of the exit is important that we get the edges that
    #   actually write the result.
    first_map_entry = state.entry_node(first_map_exit)
    first_map_scope = state.scope_subgraph(first_map_entry, include_entry=True, include_exit=True)

    # If there are more than 20 nodes we assume that the promotion is not
    #  beneficial. This is an arbitrary value.
    if first_map_scope.number_of_nodes() > 22:
        return False

    for node in first_map_scope.nodes():
        if node in (first_map_entry, first_map_exit):
            continue

        # All AccessNodes inside must be transient scalars. Because everything else
        #  indicate (empirical consideration) a more complex computation.
        if isinstance(node, dace_nodes.AccessNode):
            desc = node.desc(sdfg)
            if (not isinstance(desc, dace_data.Scalar)) and (not desc.transient):
                return False
            continue

        # A tasklet is considered to be a cheap computation.
        if isinstance(node, dace_nodes.Tasklet):
            continue

        # Everything else is considered to be a complex computation,
        #  and thus the Map should not be promoted.
        return False

    for edge in first_map_scope.edges():
        # If we have WCR then we have complex computations, most likely
        #   reductions, thus we will refuse the fusing.
        if edge.data.wcr is not None:
            return False

    return True



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import pathlib

from gt4py.cartesian.backend.module_generator import BaseModuleGenerator
from gt4py.cartesian.stencil_builder import StencilBuilder


class PythonModuleGenerator(BaseModuleGenerator):
    """Module Generator for use with backends that generate python code."""

    def __init__(self, builder: StencilBuilder) -> None:
        super().__init__(builder)

    def generate_imports(self) -> str:
        caching = self.builder.caching
        comp_pkg = f"{caching.module_prefix}computation{caching.module_postfix}"

        return "\n".join(
            [
                *super().generate_imports().splitlines(),
                "import pathlib",
                "from gt4py.cartesian.utils import make_module_from_file",
                f'computation = make_module_from_file("{comp_pkg}", pathlib.Path(__file__).parent / "{comp_pkg}.py")',
            ]
        )

    def generate_implementation(self) -> str:
        params = [f"{p.name}={p.name}" for p in self.builder.gtir.params]
        params.extend(["_domain_=_domain_", "_origin_=_origin_"])
        return f"computation.run({', '.join(params)})"


def recursive_write(root_path: pathlib.Path, tree: dict[str, str | dict]):
    root_path.mkdir(parents=True, exist_ok=True)

    for key, value in tree.items():
        if isinstance(value, dict):
            return recursive_write(root_path / key, value)

        src_path = root_path / key
        src_path.write_text(value)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.ffront import dialect_ast_enums, field_operator_ast as foast


def compute_assign_indices(
    targets: list[foast.FieldSymbol | foast.TupleSymbol | foast.ScalarSymbol | foast.Starred],
    num_elts: int,
) -> list[tuple[int, int] | int]:
    """
    Compute a list of indices, mapping each target to its respective value(s).

    This function is used when mapping a tuple of targets to a tuple of values, and also handles the
    case in which tuple unpacking is done using a Starred operator.

    Examples
    --------
    The below are examples of different types of unpacking and the generated indices. Note that the
    indices in the tuple correspond to the lower and upper slice indices of a Starred variable.

    ..  code-block:: python

        # example pseudocode
        *a, b, c = (1, 2, 3, 4)
        [(0, 2), 2, 3]

        a, *b, c = (1, 2, 3, 4)
        (0, (1, 3), 3)

        a, b, *c = (1, 2, 3, 4)
        [0, 1, (2, 4)]
    """
    starred_lower, starred_upper = None, None
    for idx, elt in enumerate(targets):
        if isinstance(elt, foast.Starred):
            starred_lower = idx
            break
    for idx, elt in zip(reversed(range(num_elts)), reversed(targets)):
        if isinstance(elt, foast.Starred):
            starred_upper = idx + 1
            break
    if starred_lower is not None and starred_upper is not None:
        return [
            *list(range(0, starred_lower)),
            (starred_lower, starred_upper),
            *list(range(starred_upper, num_elts)),
        ]
    return list(range(0, num_elts))  # no starred target


def expr_to_index(expr: foast.Expr) -> int:
    """
    Convert an expression that represent an (integral) index to an int index.

    Args:
        expr: The expression to convert. Supported are literals of the form `+/-<int>`.

    Returns:
        The integer value of the index.

    Raises:
        ValueError: If the expression is not a valid index expression.
    """
    if isinstance(expr, foast.Constant):
        return expr.value
    if (
        isinstance(expr, foast.UnaryOp)
        and isinstance(expr.op, dialect_ast_enums.UnaryOperator)
        and isinstance(expr.operand, foast.Constant)
    ):
        if expr.op is dialect_ast_enums.UnaryOperator.USUB:
            return -expr.operand.value
        if expr.op is dialect_ast_enums.UnaryOperator.UADD:
            return expr.operand.value

    raise ValueError(f"Not an index: '{expr}'.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.type_system.type_translation import (
    from_type_hint as from_type_hint,
    from_value as from_value,
    get_scalar_kind as get_scalar_kind,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Generic, Literal, Optional, TypeVar, Union

import gt4py.eve as eve
from gt4py.eve import Coerced, Node, SourceLocation, SymbolName, SymbolRef
from gt4py.eve.traits import SymbolTableTrait
from gt4py.next.ffront import dialect_ast_enums, type_specifications as ts_ffront
from gt4py.next.type_system import type_specifications as ts


class LocatedNode(Node):
    location: SourceLocation


SymbolT = TypeVar("SymbolT", bound=ts.TypeSpec)


class Symbol(eve.GenericNode, LocatedNode, Generic[SymbolT]):
    id: Coerced[SymbolName]
    type: Union[SymbolT, ts.DeferredType]  # A003
    namespace: dialect_ast_enums.Namespace = dialect_ast_enums.Namespace(
        dialect_ast_enums.Namespace.LOCAL
    )


DataTypeT = TypeVar("DataTypeT", bound=ts.DataType)
DataSymbol = Symbol[DataTypeT]

FieldTypeT = TypeVar("FieldTypeT", bound=ts.FieldType)
FieldSymbol = Symbol[FieldTypeT]

ScalarTypeT = TypeVar("ScalarTypeT", bound=ts.ScalarType)
ScalarSymbol = Symbol[ScalarTypeT]

TupleTypeT = TypeVar("TupleTypeT", bound=ts.TupleType)
TupleSymbol = Symbol[TupleTypeT]


class Expr(LocatedNode):
    type: Optional[ts.TypeSpec] = None  # A003


class BinOp(Expr):
    op: dialect_ast_enums.BinaryOperator
    left: Expr
    right: Expr


class Name(Expr):
    id: Coerced[SymbolRef]


class Call(Expr):
    func: Name
    args: list[Expr]
    kwargs: dict[str, Expr]


class Subscript(Expr):
    value: Name
    slice_: Expr


class TupleExpr(Expr):
    elts: list[Expr]


class Attribute(Expr):
    attr: str
    value: Expr


class Constant(Expr):
    value: Any  # TODO(tehrengruber): be more restrictive


class Dict(Expr):
    keys_: list[Union[Name | Attribute]]
    values_: list[TupleExpr]


class Slice(Expr):
    lower: Optional[Constant]
    upper: Optional[Constant]
    step: Literal[None]


class Stmt(LocatedNode): ...


class Program(LocatedNode, SymbolTableTrait):
    id: Coerced[SymbolName]
    type: Union[ts_ffront.ProgramType, ts.DeferredType]  # A003
    params: list[DataSymbol]
    body: list[Call]
    closure_vars: list[Symbol]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py import eve
from gt4py.cartesian.gtc.cuir import cuir


class FuseKernels(eve.NodeTranslator):
    """Fuse consecutive kernels into single kernel for better launch overhead.

    Fuse kernels that are directly translated from OIR vertical loops into separate kernels if no synchronization is required.
    """

    def visit_Program(self, node: cuir.Program) -> cuir.Program:
        def is_parallel(kernel: cuir.Kernel) -> bool:
            parallel = [
                loop.loop_order == cuir.LoopOrder.PARALLEL for loop in kernel.vertical_loops
            ]
            assert all(parallel) or not any(parallel), "Mixed k-parallelism in kernel"
            return any(parallel)

        kernels = [self.visit(node.kernels[0])]
        previous_parallel = is_parallel(kernels[-1])
        previous_writes = (
            kernels[-1]
            .walk_values()
            .if_isinstance(cuir.AssignStmt)
            .getattr("left")
            .if_isinstance(cuir.FieldAccess)
            .getattr("name")
            .to_set()
        )
        for kernel in node.kernels[1:]:
            parallel = is_parallel(kernel)
            reads_with_offsets = (
                kernel.walk_values()
                .if_isinstance(cuir.FieldAccess)
                .filter(
                    lambda x, parallel=parallel: any(
                        off != 0 for off in x.offset.to_dict().values()
                    )
                    or (x.offset.to_dict()["k"] != 0 and parallel)
                )
                .getattr("name")
                .to_set()
            )
            new_writes = (
                kernel.walk_values()
                .if_isinstance(cuir.AssignStmt)
                .getattr("left")
                .if_isinstance(cuir.FieldAccess)
                .getattr("name")
                .to_set()
            )

            if previous_parallel != parallel or reads_with_offsets & previous_writes:
                kernels.append(self.visit(kernel))
                previous_writes = new_writes
            else:
                kernels[-1].vertical_loops += self.visit(kernel.vertical_loops)
                previous_writes |= new_writes
            previous_parallel = parallel

        return cuir.Program(
            name=node.name,
            params=node.params,
            positionals=node.positionals,
            temporaries=node.temporaries,
            kernels=kernels,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import collections
import dataclasses
import enum
import functools
import operator
from typing import Optional

from gt4py import eve
from gt4py.eve import utils as eve_utils
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
    misc as ir_misc,
)
from gt4py.next.iterator.transforms import (
    fixed_point_transformation,
    inline_center_deref_lift_vars,
    inline_lambdas,
    inline_lifts,
    merge_let,
    trace_shifts,
)
from gt4py.next.iterator.type_system import inference as type_inference
from gt4py.next.type_system import type_info, type_specifications as ts


def _merge_arguments(
    args1: dict[str, itir.Expr], args2: dict[str, itir.Expr]
) -> dict[str, itir.Expr]:
    new_args = {**args1}
    for stencil_param, stencil_arg in args2.items():
        assert stencil_param not in new_args
        new_args[stencil_param] = stencil_arg
    return new_args


def _is_tuple_expr_of_literals(expr: itir.Expr):
    if cpm.is_call_to(expr, "make_tuple"):
        return all(_is_tuple_expr_of_literals(arg) for arg in expr.args)
    if cpm.is_call_to(expr, "tuple_get"):
        return _is_tuple_expr_of_literals(expr.args[1])
    return isinstance(expr, itir.Literal)


def _inline_as_fieldop_arg(
    arg: itir.Expr, *, uids: eve_utils.UIDGenerator
) -> tuple[itir.Expr, dict[str, itir.Expr]]:
    assert cpm.is_applied_as_fieldop(arg)
    arg = ir_misc.canonicalize_as_fieldop(arg)

    stencil, *_ = arg.fun.args  # type: ignore[attr-defined]  # ensured by `is_applied_as_fieldop`
    inner_args: list[itir.Expr] = arg.args
    extracted_args: dict[str, itir.Expr] = {}  # mapping from outer-stencil param to arg

    stencil_params: list[itir.Sym] = []
    stencil_body: itir.Expr = stencil.expr

    for inner_param, inner_arg in zip(stencil.params, inner_args, strict=True):
        if isinstance(inner_arg, itir.Literal):
            # note: only literals, not all scalar expressions are required as it doesn't make sense
            # for them to be computed per grid point.
            stencil_body = im.let(inner_param, im.promote_to_const_iterator(inner_arg))(
                stencil_body
            )
        else:
            stencil_params.append(inner_param)
            new_outer_stencil_param = uids.sequential_id(prefix="__iasfop")
            extracted_args[new_outer_stencil_param] = inner_arg

    return im.lift(im.lambda_(*stencil_params)(stencil_body))(
        *extracted_args.keys()
    ), extracted_args


def _deduplicate_as_fieldop_args(
    args: dict[str, itir.Expr], stencil_body: itir.Expr
) -> tuple[dict[str, itir.Expr], itir.Expr]:
    new_args_inverted: dict[itir.Expr, list[str]] = collections.defaultdict(list)
    for name, arg in args.items():
        new_args_inverted[arg].append(name)

    new_args: dict[str, itir.Expr] = {}
    new_stencil_body = stencil_body
    for arg, names in new_args_inverted.items():
        # put internal names at the end
        new_name, *aliases = sorted(names, key=lambda s: (s.startswith("_"), s))
        new_args[new_name] = arg
        if aliases:
            new_stencil_body = im.let(*((alias, new_name) for alias in aliases))(new_stencil_body)

    return new_args, new_stencil_body


def _prettify_as_fieldop_args(
    args: dict[str, itir.Expr], stencil_body: itir.Expr
) -> tuple[dict[str, itir.Expr], itir.Expr]:
    new_args: dict[str, itir.Expr] = {}
    remap_table: dict[str, str] = {}
    for name, arg in args.items():
        if isinstance(arg, itir.SymRef):
            assert arg.id not in new_args  # ensured by deduplication
            new_args[arg.id] = arg
            remap_table[name] = arg.id
        else:
            new_args[name] = arg

    return new_args, im.let(*remap_table.items())(stencil_body)


def fuse_as_fieldop(
    expr: itir.Expr, eligible_args: list[bool], *, uids: eve_utils.UIDGenerator
) -> itir.Expr:
    assert cpm.is_applied_as_fieldop(expr)

    assert isinstance(expr.fun.args[0], itir.Lambda) or cpm.is_call_to(expr.fun.args[0], "scan")
    stencil, restore_scan = ir_misc.unwrap_scan(expr.fun.args[0])

    domain = expr.fun.args[1] if len(expr.fun.args) > 1 else None

    args: list[itir.Expr] = expr.args

    new_args: dict[str, itir.Expr] = {}
    new_stencil_body: itir.Expr = stencil.expr

    for eligible, stencil_param, arg in zip(eligible_args, stencil.params, args, strict=True):
        if eligible:
            if cpm.is_applied_as_fieldop(arg):
                pass
            elif cpm.is_call_to(arg, "if_"):
                # transform scalar `if` into per-grid-point `if`
                # TODO(tehrengruber): revisit if we want to inline if_
                arg = im.op_as_fieldop("if_")(*arg.args)
            elif _is_tuple_expr_of_literals(arg):
                arg = im.op_as_fieldop(im.lambda_()(arg))()
            else:
                raise NotImplementedError()

            inline_expr, extracted_args = _inline_as_fieldop_arg(arg, uids=uids)

            new_stencil_body = im.let(stencil_param, inline_expr)(new_stencil_body)

            new_args = _merge_arguments(new_args, extracted_args)
        else:
            # just a safety check if typing information is available
            type_inference.reinfer(arg)
            if arg.type and not isinstance(arg.type, ts.DeferredType):
                assert isinstance(arg.type, ts.TypeSpec)

                assert not isinstance(
                    type_info.apply_to_primitive_constituents(type_info.extract_dtype, arg.type),
                    ts.ListType,
                )
            new_args = _merge_arguments(new_args, {stencil_param.id: arg})

    new_args, new_stencil_body = _deduplicate_as_fieldop_args(new_args, new_stencil_body)
    new_args, new_stencil_body = _prettify_as_fieldop_args(new_args, new_stencil_body)

    stencil = im.lambda_(*new_args.keys())(new_stencil_body)
    new_stencil = restore_scan(stencil)

    # simplify stencil directly to keep the tree small
    new_stencil = inline_lambdas.InlineLambdas.apply(
        new_stencil, opcount_preserving=True, force_inline_lift_args=False
    )
    new_stencil = inline_center_deref_lift_vars.InlineCenterDerefLiftVars.apply(
        new_stencil, is_stencil=True, uids=uids
    )  # to keep the tree small
    new_stencil = merge_let.MergeLet().visit(new_stencil)
    new_stencil = inline_lambdas.InlineLambdas.apply(
        new_stencil, opcount_preserving=True, force_inline_lift_args=True
    )
    new_stencil = inline_lifts.InlineLifts().visit(new_stencil)

    new_node = im.as_fieldop(new_stencil, domain)(*new_args.values())

    return new_node


def _arg_inline_predicate(node: itir.Expr, shifts: set[tuple[itir.OffsetLiteral, ...]]) -> bool:
    if _is_tuple_expr_of_literals(node):
        return True

    if (
        is_applied_fieldop := cpm.is_applied_as_fieldop(node)
        and not cpm.is_call_to(node.fun.args[0], "scan")
    ) or cpm.is_call_to(node, "if_"):
        # always inline arg if it is an applied fieldop with only a single arg
        if is_applied_fieldop and len(node.args) == 1:
            return True
        # argument is never used, will be removed when inlined
        if len(shifts) == 0:
            return True
        # applied fieldop with list return type must always be inlined as no backend supports this
        type_inference.reinfer(node)
        assert isinstance(node.type, ts.TypeSpec)
        dtype = type_info.apply_to_primitive_constituents(type_info.extract_dtype, node.type)
        if isinstance(dtype, ts.ListType):
            return True
        # only accessed at the center location
        if shifts in [set(), {()}]:
            return True
        # TODO(tehrengruber): Disabled as the InlineCenterDerefLiftVars does not support this yet
        #  and it would increase the size of the tree otherwise.
        # if len(shifts) == 1 and not any(
        #     trace_shifts.Sentinel.ALL_NEIGHBORS in access for access in shifts
        # ):
        #     return True  # noqa: ERA001 [commented-out-code]

    return False


def _make_tuple_element_inline_predicate(node: itir.Expr):
    if cpm.is_applied_as_fieldop(node):  # field, or tuple of fields
        return True
    if isinstance(node.type, ts.FieldType) and isinstance(node, itir.SymRef):
        return True
    return False


@dataclasses.dataclass(frozen=True, kw_only=True)
class FuseAsFieldOp(
    fixed_point_transformation.CombinedFixedPointTransform, eve.PreserveLocationVisitor
):
    """
    Merge multiple `as_fieldop` calls into one.

    >>> from gt4py import next as gtx
    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> IDim = gtx.Dimension("IDim")
    >>> field_type = ts.FieldType(dims=[IDim], dtype=ts.ScalarType(kind=ts.ScalarKind.INT32))
    >>> d = im.domain("cartesian_domain", {IDim: (0, 1)})
    >>> nested_as_fieldop = im.op_as_fieldop("plus", d)(
    ...     im.op_as_fieldop("multiplies", d)(
    ...         im.ref("inp1", field_type), im.ref("inp2", field_type)
    ...     ),
    ...     im.ref("inp3", field_type),
    ... )
    >>> print(nested_as_fieldop)
    as_fieldop(λ(__arg0, __arg1) → ·__arg0 + ·__arg1, c⟨ IDimₕ: [0, 1[ ⟩)(
      as_fieldop(λ(__arg0, __arg1) → ·__arg0 × ·__arg1, c⟨ IDimₕ: [0, 1[ ⟩)(inp1, inp2), inp3
    )
    >>> print(
    ...     FuseAsFieldOp.apply(
    ...         nested_as_fieldop, offset_provider_type={}, allow_undeclared_symbols=True
    ...     )
    ... )
    as_fieldop(λ(inp1, inp2, inp3) → ·inp1 × ·inp2 + ·inp3, c⟨ IDimₕ: [0, 1[ ⟩)(inp1, inp2, inp3)
    """  # noqa: RUF002  # ignore ambiguous multiplication character

    class Transformation(enum.Flag):
        #: Let `f_expr` be an expression with list dtype then
        #: `let(f, f_expr) -> as_fieldop(...)(f)` -> `as_fieldop(...)(f_expr)`
        FUSE_MAKE_TUPLE = enum.auto()
        #: `as_fieldop(...)(as_fieldop(...)(a, b), c)`
        #: -> as_fieldop(fused_stencil)(a, b, c)
        FUSE_AS_FIELDOP = enum.auto()
        INLINE_LET_VARS_OPCOUNT_PRESERVING = enum.auto()

        @classmethod
        def all(self) -> FuseAsFieldOp.Transformation:
            return functools.reduce(operator.or_, self.__members__.values())

    PRESERVED_ANNEX_ATTRS = ("domain",)
    REINFER_TYPES = True

    enabled_transformations = Transformation.all()

    uids: eve_utils.UIDGenerator

    @classmethod
    def apply(
        cls,
        node: itir.Program,
        *,
        offset_provider_type: common.OffsetProviderType,
        uids: Optional[eve_utils.UIDGenerator] = None,
        allow_undeclared_symbols=False,
        within_set_at_expr: Optional[bool] = None,
        enabled_transformations: Optional[Transformation] = None,
    ):
        enabled_transformations = enabled_transformations or cls.enabled_transformations

        node = type_inference.infer(
            node,
            offset_provider_type=offset_provider_type,
            allow_undeclared_symbols=allow_undeclared_symbols,
        )

        if within_set_at_expr is None:
            within_set_at_expr = not isinstance(node, itir.Program)

        if not uids:
            uids = eve_utils.UIDGenerator()

        new_node = cls(uids=uids, enabled_transformations=enabled_transformations).visit(
            node, within_set_at_expr=within_set_at_expr
        )
        # The `FuseAsFieldOp` pass does not fully preserve the type information yet. In particular
        # for the generated lifts this is tricky and error-prone. For simplicity, we just reinfer
        # everything here ensuring later passes can use the information.
        new_node = type_inference.infer(
            new_node,
            offset_provider_type=offset_provider_type,
            allow_undeclared_symbols=allow_undeclared_symbols,
        )
        return new_node

    def transform_fuse_make_tuple(self, node: itir.Node, **kwargs):
        if not cpm.is_call_to(node, "make_tuple"):
            return None

        for arg in node.args:
            type_inference.reinfer(arg)
            assert not isinstance(arg.type, ts.FieldType) or (
                hasattr(arg.annex, "domain")
                and isinstance(arg.annex.domain, domain_utils.SymbolicDomain)
            )

        eligible_els = [_make_tuple_element_inline_predicate(arg) for arg in node.args]
        field_args = [arg for i, arg in enumerate(node.args) if eligible_els[i]]
        distinct_domains = set(arg.annex.domain.as_expr() for arg in field_args)
        if len(distinct_domains) != len(field_args):
            new_els: list[itir.Expr | None] = [None for _ in node.args]
            field_args_by_domain: dict[itir.FunCall, list[tuple[int, itir.Expr]]] = {}
            for i, arg in enumerate(node.args):
                if eligible_els[i]:
                    assert isinstance(arg.annex.domain, domain_utils.SymbolicDomain)
                    domain = arg.annex.domain.as_expr()
                    field_args_by_domain.setdefault(domain, [])
                    field_args_by_domain[domain].append((i, arg))
                else:
                    new_els[i] = arg  # keep as is

            if len(field_args_by_domain) == 1 and all(eligible_els):
                # if we only have a single domain covering all args we don't need to create an
                # unnecessary let
                ((domain, inner_field_args),) = field_args_by_domain.items()
                new_node = im.op_as_fieldop(lambda *args: im.make_tuple(*args), domain)(
                    *(arg for _, arg in inner_field_args)
                )
                new_node = self.visit(new_node, **{**kwargs, "recurse": False})
            else:
                let_vars = {}
                for domain, inner_field_args in field_args_by_domain.items():
                    if len(inner_field_args) > 1:
                        var = self.uids.sequential_id(prefix="__fasfop")
                        fused_args = im.op_as_fieldop(lambda *args: im.make_tuple(*args), domain)(
                            *(arg for _, arg in inner_field_args)
                        )
                        type_inference.reinfer(arg)
                        # don't recurse into nested args, but only consider newly created `as_fieldop`
                        # note: this will always inline (as we inline center accessed)
                        let_vars[var] = self.visit(fused_args, **{**kwargs, "recurse": False})
                        for outer_tuple_idx, (inner_tuple_idx, _) in enumerate(inner_field_args):
                            new_el = im.tuple_get(outer_tuple_idx, var)
                            new_el.annex.domain = domain_utils.SymbolicDomain.from_expr(domain)
                            new_els[inner_tuple_idx] = new_el
                    else:
                        i, arg = inner_field_args[0]
                        new_els[i] = arg
                assert not any(el is None for el in new_els)
                assert let_vars
                new_node = im.let(*let_vars.items())(im.make_tuple(*new_els))
                new_node = inline_lambdas.inline_lambda(new_node, opcount_preserving=True)
            return new_node
        return None

    def transform_fuse_as_fieldop(self, node: itir.Node, **kwargs):
        if cpm.is_applied_as_fieldop(node):
            node = ir_misc.canonicalize_as_fieldop(node)
            stencil = node.fun.args[0]  # type: ignore[attr-defined]  # ensure cpm.is_applied_as_fieldop
            assert isinstance(stencil, itir.Lambda) or cpm.is_call_to(stencil, "scan")
            args: list[itir.Expr] = node.args
            shifts = trace_shifts.trace_stencil(stencil, num_args=len(args))

            eligible_els = [
                _arg_inline_predicate(arg, arg_shifts)
                for arg, arg_shifts in zip(args, shifts, strict=True)
            ]
            if any(eligible_els):
                return self.visit(
                    fuse_as_fieldop(node, eligible_els, uids=self.uids),
                    **{**kwargs, "recurse": False},
                )
        return None

    def transform_inline_let_vars_opcount_preserving(self, node: itir.Node, **kwargs):
        # when multiple `as_fieldop` calls are fused that use the same argument, this argument
        # might become referenced once only. In order to be able to continue fusing such arguments
        # try inlining here.
        if cpm.is_let(node):
            new_node = inline_lambdas.inline_lambda(node, opcount_preserving=True)
            if new_node is not node:  # nothing has been inlined
                return self.visit(new_node, **kwargs)

        return None

    def generic_visit(self, node, **kwargs):
        if cpm.is_applied_as_fieldop(node):  # don't descend in stencil
            return im.as_fieldop(*node.fun.args)(*self.visit(node.args, **kwargs))

        # TODO(tehrengruber): This is a common pattern that should be absorbed in
        #  `FixedPointTransformation`.
        if kwargs.get("recurse", True):
            return super().generic_visit(node, **kwargs)
        else:
            return node

    def visit(self, node, **kwargs):
        if isinstance(node, itir.SetAt):
            return itir.SetAt(
                expr=self.visit(node.expr, **kwargs | {"within_set_at_expr": True}),
                # rest doesn't need to be visited
                domain=node.domain,
                target=node.target,
            )

        # don't execute transformations unless inside `SetAt` node
        if not kwargs.get("within_set_at_expr"):
            return self.generic_visit(node, **kwargs)

        # inline all fields with list dtype. This needs to happen before the children are visited
        # such that the `as_fieldop` can be fused.
        # TODO(tehrengruber): what should we do in case the field with list dtype is a let itself?
        #  This could duplicate other expressions which we did not intend to duplicate.
        # TODO(tehrengruber): This should be moved into a `transform_` method, but
        #  `FixedPointTransformation` does not support pre-order transformations yet.
        if cpm.is_let(node):
            for arg in node.args:
                type_inference.reinfer(arg)
            eligible_els = [
                isinstance(arg.type, ts.FieldType) and isinstance(arg.type.dtype, ts.ListType)
                for arg in node.args
            ]
            if any(eligible_els):
                node = inline_lambdas.inline_lambda(node, eligible_params=eligible_els)
                return self.visit(node, **kwargs)

        node = super().visit(node, **kwargs)

        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import ast
import collections.abc
import dataclasses
import typing

from gt4py.next.ffront.fbuiltins import TYPE_BUILTIN_NAMES


ASTNodeT = typing.TypeVar("ASTNodeT", bound=ast.AST)

_UNIQUE_NAME_SEPARATOR = "ᐞ"


# TODO(tehrengruber): use a global mechanism to get uniqueness across multiple passes (not only
#  the ssa pass).
def unique_name(name: str, num_assignments: int) -> str:
    """Given the name of a variable return a unique name after the given amount of assignments."""
    if num_assignments >= 0:
        return f"{name}{_UNIQUE_NAME_SEPARATOR}{num_assignments}"
    assert num_assignments == -1
    return name


def _make_assign(target: str, source: str, location_node: ast.AST) -> ast.Assign:
    result = ast.Assign(
        targets=[ast.Name(ctx=ast.Store(), id=target)], value=ast.Name(ctx=ast.Load(), id=source)
    )
    for node in ast.walk(result):
        ast.copy_location(node, location_node)
    return result


def _is_guaranteed_to_return(node: ast.stmt | list[ast.stmt]) -> bool:
    if isinstance(node, list):
        return any(_is_guaranteed_to_return(child) for child in node)
    if isinstance(node, ast.Return):
        return True
    if isinstance(node, ast.If):
        return _is_guaranteed_to_return(node.body) and _is_guaranteed_to_return(node.orelse)
    return False


@dataclasses.dataclass
class _AssignmentTracker:
    """Helper class to keep track of the number of assignments to a variable."""

    #: mapping from a variable name to the number of times it is assigned to. `-1` signifies it to
    #:  be defined only, but not assigned to.
    _counts: dict[str, int] = dataclasses.field(default_factory=dict)

    def define(self, name: str) -> None:
        if name in self.names():
            raise ValueError(f"Variable '{name}' is already defined.")
        # -1 signifies a
        self._counts[name] = -1

    def assign(self, name: str) -> None:
        self._counts[name] = self.count(name) + 1

    def count(self, name: str) -> int:
        return self._counts.get(name, -1)

    def names(self) -> collections.abc.Set:
        return self._counts.keys()

    def copy(self) -> _AssignmentTracker:
        return _AssignmentTracker({**self._counts})


def _merge_assignment_tracker(a: _AssignmentTracker, b: _AssignmentTracker) -> _AssignmentTracker:
    return _AssignmentTracker({k: max(a.count(k), b.count(k)) for k in (a.names() & b.names())})


@dataclasses.dataclass
class SingleStaticAssignPass(ast.NodeTransformer):
    """
    Rename variables in assignments to avoid overwriting.

    Mutates the python AST, variable names will not be valid python names anymore.
    This pass must be run before any passes that linearize unpacking assignments.


    Example
    -------
    Function ``foo()`` in the following example keeps overwriting local variable ``a``

    >>> import ast, inspect

    >>> def foo():
    ...     a = 1
    ...     a = 2 + a
    ...     a = 3 + a
    ...     return a

    >>> print(ast.unparse(SingleStaticAssignPass.apply(ast.parse(inspect.getsource(foo)))))
    def foo():
        aᐞ0 = 1
        aᐞ1 = 2 + aᐞ0
        aᐞ2 = 3 + aᐞ1
        return aᐞ2

    Note that each variable name is assigned only once (per branch) and never updated / overwritten.

    Note also that after parsing, running the pass and unparsing we get invalid but
    readable python code. This is ok because this pass is not intended for
    python-to-python translation.

    WARNING: This pass is not intended as a general-purpose SSA transformation.
    The pass does not support any general Python AST. Known limitations include:
        * Nested functions aren't supported
        * While loops aren't supported
    """

    assignment_tracker: _AssignmentTracker = dataclasses.field(default_factory=_AssignmentTracker)

    @classmethod
    def apply(cls, node: ASTNodeT) -> ASTNodeT:
        return cls().visit(node)

    def _rename(self, node: ASTNodeT) -> ASTNodeT:
        for child_node in ast.walk(node):
            if isinstance(child_node, ast.Name):
                child_node.id = unique_name(
                    child_node.id, self.assignment_tracker.count(child_node.id)
                )
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        # For practical purposes, this is sufficient, but really not general at all.
        # However, the algorithm was never intended to be general.

        prev_assignment_tracker = self.assignment_tracker.copy()

        for arg in node.args.args:
            self.assignment_tracker.define(arg.arg)

        node.body = [self.visit(stmt) for stmt in node.body]

        self.assignment_tracker = prev_assignment_tracker
        return node

    def visit_If(self, node: ast.If) -> ast.If:
        prev_assignment_tracker = self.assignment_tracker

        node.test = self._rename(node.test)

        self.assignment_tracker = prev_assignment_tracker.copy()
        node.body = [self.visit(el) for el in node.body]
        body_assignment_tracker = self.assignment_tracker
        body_returns = _is_guaranteed_to_return(node.body)

        self.assignment_tracker = prev_assignment_tracker.copy()
        node.orelse = [self.visit(el) for el in node.orelse]
        orelse_assignment_tracker = self.assignment_tracker
        orelse_returns = _is_guaranteed_to_return(node.orelse)

        if body_returns and not orelse_returns:
            self.assignment_tracker = orelse_assignment_tracker
            return node

        if orelse_returns and not body_returns:
            self.assignment_tracker = body_assignment_tracker
            return node

        if body_returns and orelse_returns:
            self.assignment_tracker = _AssignmentTracker()
            return node

        assert not body_returns and not orelse_returns

        self.assignment_tracker = _merge_assignment_tracker(
            body_assignment_tracker, orelse_assignment_tracker
        )

        # ensure both branches conclude with the same unique names
        for name in self.assignment_tracker.names():
            assignment_count = self.assignment_tracker.count(name)
            body_assignment_count = body_assignment_tracker.count(name)
            orelse_assignment_count = orelse_assignment_tracker.count(name)

            if body_assignment_count != assignment_count:
                new_assign = _make_assign(
                    unique_name(name, assignment_count),
                    unique_name(name, body_assignment_count),
                    node,
                )
                node.body.append(new_assign)
            elif orelse_assignment_count != assignment_count:
                new_assign = _make_assign(
                    unique_name(name, assignment_count),
                    unique_name(name, orelse_assignment_count),
                    node,
                )
                node.orelse.append(new_assign)

        return node

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        # first update rhs names to reference the latest version
        node.value = self._rename(node.value)
        # then update lhs to create new names
        node.targets = [self.visit(target) for target in node.targets]
        return node

    def visit_Return(self, node: ast.Return) -> ast.Return:
        node.value = self._rename(node.value) if node.value else None
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AnnAssign:
        if node.value:
            node.value = self._rename(node.value)
            node.target = self.visit(node.target)
        elif isinstance(node.target, ast.Name):
            # An annotated assignment without a value always applies to the next assignment. Ensure
            # the two have the same name by keeping the assignment tracker as is (after this
            # statement).
            prev_assignment_tracker = self.assignment_tracker.copy()
            node.target = self.visit(node.target)
            self.assignment_tracker = prev_assignment_tracker
        return node

    def visit_Name(self, node: ast.Name) -> ast.Name:
        if node.id in TYPE_BUILTIN_NAMES:
            return node

        self.assignment_tracker.assign(node.id)
        node.id = unique_name(node.id, self.assignment_tracker.count(node.id))
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import ast
import textwrap
import typing
from dataclasses import dataclass
from typing import Callable

from gt4py.eve.concepts import SourceLocation
from gt4py.eve.extended_typing import Any, Generic, TypeVar
from gt4py.next import errors
from gt4py.next.ffront.ast_passes.fix_missing_locations import FixMissingLocations
from gt4py.next.ffront.ast_passes.remove_docstrings import RemoveDocstrings
from gt4py.next.ffront.source_utils import SourceDefinition, get_closure_vars_from_function


DialectRootT = TypeVar("DialectRootT")


def parse_source_definition(source_definition: SourceDefinition) -> ast.AST:
    try:
        return ast.parse(textwrap.dedent(source_definition.source)).body[0]
    except SyntaxError as err:
        assert err.lineno
        assert err.offset
        loc = SourceLocation(
            line=err.lineno + source_definition.line_offset,
            column=err.offset + source_definition.column_offset,
            filename=source_definition.filename,
            end_line=(
                err.end_lineno + source_definition.line_offset
                if err.end_lineno is not None
                else None
            ),
            end_column=(
                err.end_offset + source_definition.column_offset
                if err.end_offset is not None
                else None
            ),
        )
        raise errors.DSLError(loc, err.msg).with_traceback(err.__traceback__) from err


@dataclass(frozen=True, kw_only=True)
class DialectParser(ast.NodeVisitor, Generic[DialectRootT]):
    source_definition: SourceDefinition
    closure_vars: dict[str, Any]
    annotations: dict[str, Any]

    @classmethod
    def apply(
        cls,
        source_definition: SourceDefinition,
        closure_vars: dict[str, Any],
        annotations: dict[str, Any],
    ) -> DialectRootT:
        definition_ast: ast.AST
        definition_ast = parse_source_definition(source_definition)

        definition_ast = RemoveDocstrings.apply(definition_ast)
        definition_ast = FixMissingLocations.apply(definition_ast)
        output_ast = cls._postprocess_dialect_ast(
            cls(
                source_definition=source_definition,
                closure_vars=closure_vars,
                annotations=annotations,
            ).visit(cls._preprocess_definition_ast(definition_ast)),
            closure_vars,
            annotations,
        )

        return output_ast

    @classmethod
    def apply_to_function(cls, function: Callable) -> DialectRootT:
        src = SourceDefinition.from_function(function)
        closure_vars = get_closure_vars_from_function(function)
        annotations = typing.get_type_hints(function)
        return cls.apply(src, closure_vars, annotations)

    @classmethod
    def _preprocess_definition_ast(cls, definition_ast: ast.AST) -> ast.AST:
        return definition_ast

    @classmethod
    def _postprocess_dialect_ast(
        cls, output_ast: DialectRootT, closure_vars: dict[str, Any], annotations: dict[str, Any]
    ) -> DialectRootT:
        return output_ast

    def generic_visit(self, node: ast.AST) -> None:
        loc = self.get_location(node)
        feature = f"{type(node).__module__}.{type(node).__qualname__}"
        raise errors.UnsupportedPythonFeatureError(loc, feature)

    def get_location(self, node: ast.AST) -> SourceLocation:
        file = self.source_definition.filename
        line_offset = self.source_definition.line_offset
        col_offset = self.source_definition.column_offset

        # `FixMissingLocations` ensures that all nodes have the location attributes
        assert hasattr(node, "lineno")
        line = node.lineno + line_offset
        assert hasattr(node, "end_lineno")
        end_line = node.end_lineno + line_offset if node.end_lineno is not None else None
        assert hasattr(node, "col_offset")
        column = 1 + node.col_offset + col_offset
        assert hasattr(node, "end_col_offset")
        end_column = (
            1 + node.end_col_offset + col_offset if node.end_col_offset is not None else None
        )

        loc = SourceLocation(file, line, column, end_line=end_line, end_column=end_column)
        return loc



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from typing import Any, Callable, Optional, TypeVar

from gt4py.eve import utils as eve_utils
from gt4py.next.ffront import type_info as ti_ffront
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.type_system import type_info, type_specifications as ts


def to_tuples_of_iterator(expr: itir.Expr | str, arg_type: ts.TypeSpec) -> itir.FunCall:
    """
    Convert iterator of tuples into tuples of iterator.

    Supports arbitrary nesting.

    >>> print(
    ...     to_tuples_of_iterator(
    ...         "arg",
    ...         ts.TupleType(
    ...             types=[ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT32))]
    ...         ),
    ...     )
    ... )  # doctest: +ELLIPSIS
    (λ(__toi_...) → {(↑(λ(it) → (·it)[0]))(__toi_...)})(arg)
    """
    param = f"__toi_{eve_utils.content_hash(expr)}"

    def fun(primitive_type: ts.TypeSpec, path: tuple[int, ...]) -> itir.Expr:
        inner_expr = im.deref("it")
        for path_part in path:
            inner_expr = im.tuple_get(path_part, inner_expr)

        return im.lift(im.lambda_("it")(inner_expr))(param)

    return im.let(param, expr)(
        type_info.apply_to_primitive_constituents(
            fun, arg_type, with_path_arg=True, tuple_constructor=im.make_tuple
        )
    )


def to_iterator_of_tuples(expr: itir.Expr | str, arg_type: ts.TypeSpec) -> itir.FunCall:
    """
    Convert tuples of iterator into iterator of tuples.

    Supports arbitrary nesting.

    >>> print(
    ...     to_iterator_of_tuples(
    ...         "arg",
    ...         ts.TupleType(
    ...             types=[ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT32))]
    ...         ),
    ...     )
    ... )  # doctest: +ELLIPSIS
    (λ(__iot_...) → (↑(λ(__iot_el_0) → {·__iot_el_0}))(__iot_...[0]))(arg)
    """
    param = f"__iot_{eve_utils.content_hash(expr)}"

    type_constituents = [
        ti_ffront.promote_scalars_to_zero_dim_field(type_)
        for type_ in type_info.primitive_constituents(arg_type)
    ]
    assert all(
        isinstance(type_, ts.FieldType) and type_.dims == type_constituents[0].dims  # type: ignore[attr-defined]  # ensure by assert above
        for type_ in type_constituents
    )

    def fun(_: Any, path: tuple[int, ...]) -> itir.FunCall:
        param_name = "__iot_el"
        for path_part in path:
            param_name = f"{param_name}_{path_part}"
        return im.deref(param_name)

    lift_params, lift_args = [], []
    for _, path in type_info.primitive_constituents(arg_type, with_path_arg=True):
        param_name, arg_expr = "__iot_el", param
        for path_part in path:
            param_name = f"{param_name}_{path_part}"
            arg_expr = im.tuple_get(path_part, arg_expr)

        lift_params.append(param_name)
        lift_args.append(arg_expr)

    stencil_expr = type_info.apply_to_primitive_constituents(
        fun, arg_type, with_path_arg=True, tuple_constructor=im.make_tuple
    )
    return im.let(param, expr)(im.lift(im.lambda_(*lift_params)(stencil_expr))(*lift_args))


# TODO(tehrengruber): The code quality of this function is poor. We should rewrite it.
def process_elements(
    process_func: Callable[..., itir.Expr],
    objs: itir.Expr | Iterable[itir.Expr],
    current_el_type: ts.TypeSpec,
    arg_types: Optional[Iterable[ts.TypeSpec]] = None,
) -> itir.FunCall:
    """
    Recursively applies a processing function to all primitive constituents of a tuple.

    Arguments:
        process_func: A callable that takes an itir.Expr representing a leaf-element of `objs`.
            If multiple `objs` are given the callable takes equally many arguments.
        objs: The object whose elements are to be transformed.
        current_el_type: A type with the same structure as the elements of `objs`. The leaf-types
            are not used and thus not relevant.
        arg_types: If provided, a tuple of the type of each argument is passed to `process_func` as last argument.
            Note, that `arg_types` might coincide with `(current_el_type,)*len(objs)`, but not necessarily,
            in case of implicit broadcasts.
    """
    if isinstance(objs, itir.Expr):
        objs = (objs,)

    let_ids = tuple(f"__val_{eve_utils.content_hash(obj)}" for obj in objs)
    body = _process_elements_impl(
        process_func,
        tuple(im.ref(let_id) for let_id in let_ids),
        current_el_type,
        arg_types=arg_types,
    )

    return im.let(*(zip(let_ids, objs, strict=True)))(body)


T = TypeVar("T", bound=itir.Expr, covariant=True)


def _process_elements_impl(
    process_func: Callable[..., itir.Expr],
    _current_el_exprs: Iterable[T],
    current_el_type: ts.TypeSpec,
    arg_types: Optional[Iterable[ts.TypeSpec]],
) -> itir.Expr:
    if isinstance(current_el_type, ts.TupleType):
        result = im.make_tuple(
            *(
                _process_elements_impl(
                    process_func,
                    tuple(
                        im.tuple_get(i, current_el_expr) for current_el_expr in _current_el_exprs
                    ),
                    current_el_type.types[i],
                    arg_types=tuple(arg_t.types[i] for arg_t in arg_types)  # type: ignore[attr-defined] # guaranteed by the requirement that `current_el_type` and each element of `arg_types` have the same tuple structure
                    if arg_types is not None
                    else None,
                )
                for i in range(len(current_el_type.types))
            )
        )
    else:
        if arg_types is not None:
            result = process_func(*_current_el_exprs, arg_types)
        else:
            result = process_func(*_current_el_exprs)

    return result



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Any, Optional

import dace
import factory

from gt4py._core import definitions as core_defs
from gt4py.next import common, config, metrics
from gt4py.next.iterator import ir as itir, transforms as itir_transforms
from gt4py.next.otf import arguments, languages, stages, step_types, workflow
from gt4py.next.otf.binding import interface
from gt4py.next.otf.languages import LanguageSettings
from gt4py.next.program_processors.runners.dace import (
    gtir_to_sdfg,
    transformations as gtx_transformations,
    utils as gtx_dace_utils,
)
from gt4py.next.program_processors.runners.dace.workflow import common as gtx_wfcommon
from gt4py.next.type_system import type_specifications as ts


def find_constant_symbols(
    ir: itir.Program,
    sdfg: dace.SDFG,
    offset_provider_type: common.OffsetProviderType,
) -> dict[str, int]:
    """Helper function to find symbols to replace with constant values."""
    constant_symbols: dict[str, int] = {}
    if config.UNSTRUCTURED_HORIZONTAL_HAS_UNIT_STRIDE:
        # Search the stride symbols corresponding to the horizontal dimension
        for p in ir.params:
            if isinstance(p.type, ts.FieldType):
                dims = p.type.dims
                if len(dims) == 0:
                    continue
                elif len(dims) == 1:
                    dim_index = 0
                elif len(dims) == 2:
                    dim_index = 0 if dims[0].kind == common.DimensionKind.HORIZONTAL else 1
                else:
                    raise ValueError(f"Unsupported field with dims={dims}.")
                stride_name = gtx_dace_utils.field_stride_symbol_name(p.id, dim_index)
                constant_symbols[stride_name] = 1
        # Same for connectivity tables, for which the first dimension is always horizontal
        for conn, desc in sdfg.arrays.items():
            if gtx_dace_utils.is_connectivity_identifier(conn, offset_provider_type):
                assert not desc.transient
                stride_name = gtx_dace_utils.field_stride_symbol_name(conn, 0)
                constant_symbols[stride_name] = 1

    return constant_symbols


def make_sdfg_call_async(sdfg: dace.SDFG, gpu: bool) -> None:
    """Configure an SDFG to immediately return once all work has been scheduled.

    This means that `CompiledSDFG.fast_call()` will return immediately after all
    computations have been _scheduled_ on the device. This function only has an effect
    for work that runs on the GPU. Furthermore, all work is scheduled on the
    default stream.

    Todo: Revisit this function once DaCe changes its behaviour in this regard.
    """

    # This is only a problem on GPU.
    # TODO(phimuell): Figuring out what about OpenMP.
    if not gpu:
        return

    assert dace.Config.get("compiler.cuda.max_concurrent_streams") == -1, (
        f"Expected `max_concurrent_streams == -1` but it was `{dace.Config.get('compiler.cuda.max_concurrent_streams')}`."
    )

    # NOTE: We are using the default stream this means that _**currently**_ the launch is
    #   already asynchronous, see [DaCe issue#2120](https://github.com/spcl/dace/issues/2120)
    #   for more. However, DaCe still [generates streams internally](https://github.com/spcl/dace/blob/54c935cfe74a52c5107dc91680e6201ddbf86821/dace/codegen/targets/cuda.py#L467).
    #   Thus to be absolutely sure we will no set all streams, DaCe uses to the default
    #   stream.
    # NOTE: Another important note here is, that it looks as if no synchronization
    #   what soever between states is generated if the default stream is used. This
    #   might lead to problems if a GPU kernel computes something that is needed for a
    #   condition of an interstate edge. However, we should not have that case.
    # TODO(phimuell, edopao): Make sure if this is really the case.
    dace_gpu_backend = dace.Config.get("compiler.cuda.backend")
    assert dace_gpu_backend in ["cuda", "hip"], f"GPU backend '{dace_gpu_backend}' is unknown."
    sdfg.append_init_code(
        f"__dace_gpu_set_all_streams(__state, {dace_gpu_backend}StreamDefault);",
        location="cuda",
    )


def make_sdfg_call_sync(sdfg: dace.SDFG, gpu: bool) -> None:
    """Process the SDFG such that the call is synchronous.

    This means that `CompiledSDFG.fast_call()` will return only after all computations
    have _finished_ and the results are available. This function only has an effect for
    work that runs on the GPU. Furthermore, all work is scheduled on the default stream.

    Todo: Revisit this function once DaCe changes its behaviour in this regard.
    """

    # This is only a problem on GPU.
    # TODO(phimuell): Figuring out what about OpenMP.
    if not gpu:
        return

    assert dace.Config.get("compiler.cuda.max_concurrent_streams") == -1, (
        f"Expected `max_concurrent_streams == -1` but it was `{dace.Config.get('compiler.cuda.max_concurrent_streams')}`."
    )

    # If we are using the default stream, things are a bit simpler/harder. For some
    #  reasons when using the default stream, DaCe seems to skip _all_ synchronization,
    #  for more see [DaCe issue#2120](https://github.com/spcl/dace/issues/2120).
    #  Thus the `CompiledSDFG.fast_call()` call is truly asynchronous, i.e. just
    #  launches the kernels and then exist. Thus we have to add a synchronization
    #  at the end to have a synchronous call. We can not use `SDFG.append_exit_code()`
    #  because that code is only run at the `exit()` stage, not after a call. Thus we
    #  will generate an SDFGState that contains a Tasklet with the sync call.
    sync_state = sdfg.add_state("sync_state")
    for sink_state in sdfg.sink_nodes():
        if sink_state is sync_state:
            continue
        sdfg.add_edge(sink_state, sync_state, dace.InterstateEdge())
    assert sdfg.in_degree(sync_state) > 0

    # NOTE: Since the synchronization is done through the Tasklet explicitly,
    #   we can disable synchronization for the last state. Might be useless.
    sync_state.nosync = True

    # NOTE: We should actually wrap the `StreamSynchronize` function inside a
    #   `DACE_GPU_CHECK()` macro. However, this only works in GPU context, but
    #   here we are in CPU context. Thus we can not do it.
    dace_gpu_backend = dace.Config.get("compiler.cuda.backend")
    assert dace_gpu_backend in ["cuda", "hip"], f"GPU backend '{dace_gpu_backend}' is unknown."
    sync_state.add_tasklet(
        "sync_tlet",
        inputs=set(),
        outputs=set(),
        code=f"{dace_gpu_backend}StreamSynchronize({dace_gpu_backend}StreamDefault);",
        language=dace.dtypes.Language.CPP,
        side_effects=True,
    )

    # DaCe [still generates a stream](https://github.com/spcl/dace/blob/54c935cfe74a52c5107dc91680e6201ddbf86821/dace/codegen/targets/cuda.py#L467)
    #  despite not using it. Thus to be absolutely sure, we will not set that stream
    #  to the default stream.
    sdfg.append_init_code(
        f"__dace_gpu_set_all_streams(__state, {dace_gpu_backend}StreamDefault);",
        location="cuda",
    )


@dataclasses.dataclass(frozen=True)
class DaCeTranslator(
    workflow.ChainableWorkflowMixin[
        stages.CompilableProgram, stages.ProgramSource[languages.SDFG, languages.LanguageSettings]
    ],
    step_types.TranslationStep[languages.SDFG, languages.LanguageSettings],
):
    device_type: core_defs.DeviceType
    auto_optimize: bool
    async_sdfg_call: bool = False
    disable_itir_transforms: bool = False
    disable_field_origin_on_program_arguments: bool = False

    # auto-optimize arguments
    gpu_block_size: tuple[int, int, int] = (32, 8, 1)
    make_persistent: bool = False
    use_memory_pool: bool = False
    blocking_dim: Optional[common.Dimension] = None
    blocking_size: int = 10
    validate: bool = False
    validate_all: bool = False

    def generate_sdfg(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> dace.SDFG:
        # NOTE: We should actually create a new configuration context here, such that
        #   we do not influence anyone else. However, because [DaCe issue#2125](https://github.com/spcl/dace/issues/2125)
        #   we can not do this. Instead we simply set the configuration and hope
        #   for the best.
        gtx_wfcommon.set_dace_config(device_type=self.device_type)
        return self._generate_sdfg_without_configuring_dace(*args, **kwargs)

    def _generate_sdfg_without_configuring_dace(
        self,
        ir: itir.Program,
        offset_provider: common.OffsetProvider,
        column_axis: Optional[common.Dimension],
    ) -> dace.SDFG:
        if not self.disable_itir_transforms:
            ir = itir_transforms.apply_fieldview_transforms(ir, offset_provider=offset_provider)
        offset_provider_type = common.offset_provider_to_type(offset_provider)
        on_gpu = self.device_type != core_defs.DeviceType.CPU

        if self.use_memory_pool and not on_gpu:
            raise NotImplementedError("Memory pool only available for GPU device.")

        sdfg = gtir_to_sdfg.build_sdfg_from_gtir(
            ir,
            offset_provider_type,
            column_axis,
            disable_field_origin_on_program_arguments=self.disable_field_origin_on_program_arguments,
        )

        if self.auto_optimize:
            unit_strides_kind = (
                common.DimensionKind.HORIZONTAL
                if config.UNSTRUCTURED_HORIZONTAL_HAS_UNIT_STRIDE
                else None  # let `gt_auto_optimize` select `unit_strides_kind` based on `gpu` argument
            )
            constant_symbols = find_constant_symbols(ir, sdfg, offset_provider_type)
            gtx_transformations.gt_auto_optimize(
                sdfg,
                gpu=on_gpu,
                gpu_block_size=self.gpu_block_size,
                unit_strides_kind=unit_strides_kind,
                constant_symbols=constant_symbols,
                assume_pointwise=True,
                make_persistent=self.make_persistent,
                gpu_memory_pool=self.use_memory_pool,
                blocking_dim=self.blocking_dim,
                blocking_size=self.blocking_size,
                validate=self.validate,
                validate_all=self.validate_all,
            )
        elif on_gpu:
            # We run simplify to bring the SDFG into a canonical form that the GPU transformations
            # can handle. This is a workaround for an issue with scalar expressions that are
            # promoted to symbolic expressions and computed on the host (CPU), but the intermediate
            # result is written to a GPU global variable (https://github.com/spcl/dace/issues/1773).
            gtx_transformations.gt_simplify(sdfg)
            gtx_transformations.gt_gpu_transformation(sdfg, try_removing_trivial_maps=True)

        async_sdfg_call = False
        if config.COLLECT_METRICS_LEVEL != metrics.DISABLED:
            # We measure the execution time of one program by instrumenting the
            #   top-level SDFG with a cpp timer (std::chrono). This timer measures
            #   only the computation time, it does not include the overhead of
            #   calling the SDFG from Python.
            sdfg.instrument = dace.dtypes.InstrumentationType.Timer
        elif self.async_sdfg_call:
            async_sdfg_call = True

        if async_sdfg_call:
            make_sdfg_call_async(sdfg, on_gpu)
        else:
            make_sdfg_call_sync(sdfg, on_gpu)

        return sdfg

    def __call__(
        self, inp: stages.CompilableProgram
    ) -> stages.ProgramSource[languages.SDFG, LanguageSettings]:
        """Generate DaCe SDFG file from the GTIR definition."""
        program: itir.Program = inp.data
        assert isinstance(program, itir.Program)

        sdfg = self.generate_sdfg(
            program,
            inp.args.offset_provider,  # TODO(havogt): should be offset_provider_type once the transformation don't require run-time info
            inp.args.column_axis,
        )

        arg_types = tuple(
            arg.type_ if isinstance(arg, arguments.StaticArg) else arg for arg in inp.args.args
        )

        program_parameters = tuple(
            interface.Parameter(param.id, arg_type)
            for param, arg_type in zip(program.params, arg_types)
        )

        module: stages.ProgramSource[languages.SDFG, languages.LanguageSettings] = (
            stages.ProgramSource(
                entry_point=interface.Function(program.id, program_parameters),
                # Set 'hash=True' to compute the SDFG hash and store it in the JSON.
                #   We compute the hash in order to refresh `cfg_list` on the SDFG,
                #   which makes the JSON serialization stable.
                source_code=sdfg.to_json(hash=True),
                library_deps=tuple(),
                language=languages.SDFG,
                language_settings=languages.LanguageSettings(
                    formatter_key="", formatter_style="", file_extension="sdfg"
                ),
                implicit_domain=inp.data.implicit_domain,
            )
        )
        return module


class DaCeTranslationStepFactory(factory.Factory):
    class Meta:
        model = DaCeTranslator



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implements the lowering of concat_where operator.

This builtin translator implements the `PrimitiveTranslator` protocol as other
translators in `gtir_to_sdfg_primitives` module.
"""

from __future__ import annotations

import dace
import sympy
from dace import subsets as dace_subsets

from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm
from gt4py.next.program_processors.runners.dace import (
    gtir_domain,
    gtir_to_sdfg,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
)
from gt4py.next.type_system import type_specifications as ts


def _make_concat_field_slice(
    ctx: gtir_to_sdfg.SubgraphContext,
    field: gtir_to_sdfg_types.FieldopData,
    field_desc: dace.data.Array,
    concat_dim: gtx_common.Dimension,
    concat_dim_index: int,
    concat_dim_origin: dace.symbolic.SymbolicType,
) -> tuple[gtir_to_sdfg_types.FieldopData, dace.data.Array]:
    """
    Helper function called by `_translate_concat_where_impl` to create a slice along
    the concat dimension, that is a new array with an extra dimension and a single
    level. This allows to concatanate the input fields along the concat dimension.
    """
    assert isinstance(field.gt_type, ts.FieldType)
    assert concat_dim not in field.gt_type.dims
    dims = [
        *field.gt_type.dims[:concat_dim_index],
        concat_dim,
        *field.gt_type.dims[concat_dim_index:],
    ]
    origin = tuple(
        [*field.origin[:concat_dim_index], concat_dim_origin, *field.origin[concat_dim_index:]]
    )
    shape = tuple([*field_desc.shape[:concat_dim_index], 1, *field_desc.shape[concat_dim_index:]])
    extended_field_data, extended_field_desc = ctx.sdfg.add_temp_transient(shape, field_desc.dtype)
    extended_field_node = ctx.state.add_access(extended_field_data)
    ctx.state.add_nedge(
        field.dc_node,
        extended_field_node,
        dace.Memlet(
            data=field.dc_node.data,
            subset=dace_subsets.Range.from_array(field_desc),
            other_subset=dace_subsets.Range.from_array(extended_field_desc),
        ),
    )
    extended_field = gtir_to_sdfg_types.FieldopData(
        extended_field_node, ts.FieldType(dims=dims, dtype=field.gt_type.dtype), origin
    )
    return extended_field, extended_field_desc


def _make_concat_scalar_broadcast(
    ctx: gtir_to_sdfg.SubgraphContext,
    inp: gtir_to_sdfg_types.FieldopData,
    inp_desc: dace.data.Array,
    out_domain: gtir_domain.FieldopDomain,
    concat_dim_index: int,
) -> tuple[gtir_to_sdfg_types.FieldopData, dace.data.Array]:
    """
    Helper function called by `_translate_concat_where_impl` to create a mapped
    tasklet that broadcasts one scalar value on the given domain.

    The scalar value can come from either a scalar node or from a 1D-array (assuming
    the array represents a field in the concat dimension).
    """
    assert isinstance(inp.gt_type, ts.FieldType)
    assert len(inp.gt_type.dims) == 1
    out_dims, out_origin, out_shape = gtir_domain.get_field_layout(out_domain, ctx.target_domain)
    out_type = ts.FieldType(dims=out_dims, dtype=inp.gt_type.dtype)

    out_name, out_desc = ctx.sdfg.add_temp_transient(out_shape, inp_desc.dtype)
    out_node = ctx.state.add_access(out_name)

    map_variables = [gtir_to_sdfg_utils.get_map_variable(dim) for dim in out_dims]
    inp_index = (
        "0"
        if isinstance(inp.dc_node.desc(ctx.sdfg), dace.data.Scalar)
        else (
            f"({map_variables[concat_dim_index]} + {out_origin[concat_dim_index] - inp.origin[0]})"
        )
    )
    ctx.state.add_mapped_tasklet(
        "broadcast",
        map_ranges=dict(zip(map_variables, dace_subsets.Range.from_array(out_desc), strict=True)),
        code="__out = __inp",
        inputs={"__inp": dace.Memlet(data=inp.dc_node.data, subset=inp_index)},
        outputs={"__out": dace.Memlet(data=out_name, subset=",".join(map_variables))},
        input_nodes={inp.dc_node},
        output_nodes={out_node},
        external_edges=True,
    )

    out_field = gtir_to_sdfg_types.FieldopData(out_node, out_type, tuple(out_origin))
    return out_field, out_desc


def _translate_concat_where_impl(
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    mask_domain: gtir_domain.FieldopDomain,
    node_domain: gtir.Expr,
    tb_node_domain: gtir.Expr,
    fb_node_domain: gtir.Expr,
    tb_field: gtir_to_sdfg_types.FieldopData,
    fb_field: gtir_to_sdfg_types.FieldopData,
) -> gtir_to_sdfg_types.FieldopData:
    """
    Helper function called by `translate_concat_where()` to lower 'concat_where'
    on a single output field.

    In case of tuples, this function is called on all fields by means of `tree_map`.

    It builds the output field by concatanating the two input fields on the lower
    and upper domain. These two domains are computed from the intersection of the
    mask and the input domains.

    Note that 'tb' and 'fb' stand for true/false branch and refer to the two branches
    of the concat_where expression, passed as second and third argument, respectively.

    Args:
        ctx: The SDFG context where the primitive subgraph should be instantiated.
        sdfg_builder: The object responsible for SubgraphContexting child nodes of the primitive node.
        mask_domain: Domain (only for concat dimension) of the true branch, infinite
            on lower or upper boundary.
        node_domain: Domain (all dimensions) of output field.
        tb_node_domain: Domain of the field passed on the true branch.
        fb_node_domain: Domain of the field passed on the false branch.
        tb_field: Input field on the true branch.
        fb_field: Input field on the false branch.

    Returns:
        The field resulted from concatanating the input fields on the lower and upper domain.
    """
    tb_data_desc, fb_data_desc = (inp.dc_node.desc(ctx.sdfg) for inp in [tb_field, fb_field])
    assert tb_data_desc.dtype == fb_data_desc.dtype

    tb_domain, fb_domain = (
        gtir_domain.extract_domain(domain) for domain in [tb_node_domain, fb_node_domain]
    )
    assert len(mask_domain) == 1
    concat_domain = mask_domain[0]

    # Expect unbound range in the concat domain expression on lower or upper range:
    #  - if the domain expression is unbound on lower side (negative infinite),
    #    the expression on the true branch is to be considered the input for the
    #    lower domain.
    #  - viceversa, if the domain expression is unbound on upper side (positive
    #    infinite), the true expression represents the input for the upper domain.
    if concat_domain.start == gtir_to_sdfg_utils.get_symbolic(gtir.InfinityLiteral.NEGATIVE):
        concat_dim_bound = concat_domain.stop
        lower, lower_desc, lower_domain = (tb_field, tb_data_desc, tb_domain)
        upper, upper_desc, upper_domain = (fb_field, fb_data_desc, fb_domain)
    elif concat_domain.stop == gtir_to_sdfg_utils.get_symbolic(gtir.InfinityLiteral.POSITIVE):
        concat_dim_bound = concat_domain.start
        lower, lower_desc, lower_domain = (fb_field, fb_data_desc, fb_domain)
        upper, upper_desc, upper_domain = (tb_field, tb_data_desc, tb_domain)
    else:
        raise ValueError(f"Unexpected concat mask {mask_domain[0]}.")

    # we use the concat domain, stored in the annex, as the domain of output field
    output_domain = gtir_domain.extract_domain(node_domain)
    output_dims, output_origin, output_shape = gtir_domain.get_field_layout(
        output_domain, ctx.target_domain
    )
    concat_dim_index = output_dims.index(concat_domain.dim)

    """
    In case one of the arguments is a scalar value, for example:
    ```python
    @gtx.field_operator
    def testee(a: np.int32, b: cases.IJKField) -> cases.IJKField:
        return concat_where(KDim < 1, a, b)
    ```
    we convert it to a single-element 1D field with the dimension of the concat expression.
    """
    if isinstance(lower.gt_type, ts.ScalarType):
        assert len(lower_domain) == 0
        assert isinstance(upper.gt_type, ts.FieldType)
        lower = gtir_to_sdfg_types.FieldopData(
            lower.dc_node,
            ts.FieldType(dims=[concat_domain.dim], dtype=lower.gt_type),
            origin=(concat_dim_bound - 1,),
        )
        lower_domain = [
            gtir_domain.FieldopDomainRange(
                dim=concat_domain.dim,
                start=output_domain[concat_dim_index].start,
                stop=concat_dim_bound,
            )
        ]
    elif isinstance(upper.gt_type, ts.ScalarType):
        assert len(upper_domain) == 0
        assert isinstance(lower.gt_type, ts.FieldType)
        upper = gtir_to_sdfg_types.FieldopData(
            upper.dc_node,
            ts.FieldType(dims=[concat_domain.dim], dtype=upper.gt_type),
            origin=(concat_dim_bound,),
        )
        upper_domain = [
            gtir_domain.FieldopDomainRange(
                dim=concat_domain.dim,
                start=concat_dim_bound,
                stop=output_domain[concat_dim_index].stop,
            )
        ]

    if concat_domain.dim not in lower.gt_type.dims:  # type: ignore[union-attr]
        """
        The field on the lower domain is to be treated as a slice to add as one
        level in the concat dimension, on the lower bound.
        Consider for example the following IR, where a horizontal field is added
        as level zero in K-dimension:
        ```python
        @gtx.field_operator
        def testee(interior: cases.IJKField, boundary: cases.IJField) -> cases.IJKField:
            return concat_where(KDim == 0, boundary, interior)
        ```
        """
        assert (
            lower.gt_type.dims  # type: ignore[union-attr]
            == [
                *upper.gt_type.dims[0:concat_dim_index],  # type: ignore[union-attr]
                *upper.gt_type.dims[concat_dim_index + 1 :],  # type: ignore[union-attr]
            ]
        )
        lower, lower_desc = _make_concat_field_slice(
            ctx=ctx,
            field=lower,
            field_desc=lower_desc,
            concat_dim=concat_domain.dim,
            concat_dim_index=concat_dim_index,
            concat_dim_origin=concat_dim_bound - 1,
        )
        lower_domain.insert(
            concat_dim_index,
            gtir_domain.FieldopDomainRange(
                dim=concat_domain.dim,
                start=gtir_domain.simplify_domain_expr(
                    sympy.Max(concat_dim_bound - 1, output_domain[concat_dim_index].start),
                    ctx.target_domain,
                ),
                stop=concat_dim_bound,
            ),
        )
    elif concat_domain.dim not in upper.gt_type.dims:  # type: ignore[union-attr]
        # Same as previous case, but the field slice is added on the upper bound.
        assert (
            upper.gt_type.dims  # type: ignore[union-attr]
            == [
                *lower.gt_type.dims[0:concat_dim_index],  # type: ignore[union-attr]
                *lower.gt_type.dims[concat_dim_index + 1 :],  # type: ignore[union-attr]
            ]
        )
        upper, upper_desc = _make_concat_field_slice(
            ctx=ctx,
            field=upper,
            field_desc=upper_desc,
            concat_dim=concat_domain.dim,
            concat_dim_index=concat_dim_index,
            concat_dim_origin=concat_dim_bound,
        )
        upper_domain.insert(
            concat_dim_index,
            gtir_domain.FieldopDomainRange(
                dim=concat_domain.dim,
                start=concat_dim_bound,
                stop=gtir_domain.simplify_domain_expr(
                    sympy.Min(concat_dim_bound + 1, output_domain[concat_dim_index].stop),
                    ctx.target_domain,
                ),
            ),
        )
    elif isinstance(lower_desc, dace.data.Scalar) or (
        len(lower.gt_type.dims) == 1 and len(output_domain) > 1  # type: ignore[union-attr]
    ):
        """
        The input on the lower domain is either a scalar or a 1d field, representing
        the value(s) to be added as one level in the concat dimension below the upper domain.
        Consider for example the following IR, where the scalar value is one level
        (`KDim == 0`) taken from lower input 'a':
        ```python
        @gtx.field_operator
        def testee(a: cases.KField, b: cases.IJKField) -> cases.IJKField:
            return concat_where(KDim == 0, a, b)
        ```
        """
        assert len(lower_domain) == 1 and lower_domain[0].dim == concat_domain.dim
        lower_domain = [
            *output_domain[:concat_dim_index],
            lower_domain[0],
            *output_domain[concat_dim_index + 1 :],
        ]
        lower, lower_desc = _make_concat_scalar_broadcast(
            ctx=ctx,
            inp=lower,
            inp_desc=lower_desc,
            out_domain=lower_domain,
            concat_dim_index=concat_dim_index,
        )
    elif isinstance(upper_desc, dace.data.Scalar) or (
        len(upper.gt_type.dims) == 1 and len(output_domain) > 1  # type: ignore[union-attr]
    ):
        # Same as previous case, but the scalar value is taken from `upper` input.
        assert len(upper_domain) == 1 and upper_domain[0].dim == concat_domain.dim
        upper_domain = [
            *output_domain[:concat_dim_index],
            upper_domain[0],
            *output_domain[concat_dim_index + 1 :],
        ]
        upper, upper_desc = _make_concat_scalar_broadcast(
            ctx=ctx,
            inp=upper,
            inp_desc=upper_desc,
            out_domain=upper_domain,
            concat_dim_index=concat_dim_index,
        )
    else:
        """
        Handle here the _regular_ case, that is concat_where applied to two fields
        with same domain:
        ```python
        @gtx.field_operator
        def testee(a: cases.IJKField, b: cases.IJKField) -> cases.IJKField:
            return concat_where(KDim <=10 , a, b)
        ```
        """
        assert isinstance(lower.gt_type, ts.FieldType)
        assert isinstance(lower_desc, dace.data.Array)
        assert isinstance(upper.gt_type, ts.FieldType)
        assert isinstance(upper_desc, dace.data.Array)
        if lower.gt_type.dims != upper.gt_type.dims:
            raise NotImplementedError(
                "Lowering concat_where on fields with different domain is not supported."
            )

    # ensure that the arguments have the same domain as the concat result
    assert all(ftype.dims == output_dims for ftype in (lower.gt_type, upper.gt_type))  # type: ignore[union-attr]

    lower_range_0 = output_domain[concat_dim_index].start
    lower_range_1 = gtir_domain.simplify_domain_expr(
        sympy.Max(lower_range_0, lower_domain[concat_dim_index].stop),
        ctx.target_domain,
    )
    lower_range_size = lower_range_1 - lower_range_0

    upper_range_1 = output_domain[concat_dim_index].stop
    upper_range_0 = gtir_domain.simplify_domain_expr(
        sympy.Min(upper_range_1, upper_domain[concat_dim_index].start),
        ctx.target_domain,
    )
    upper_range_size = upper_range_1 - upper_range_0

    output, output_desc = sdfg_builder.add_temp_array(ctx.sdfg, output_shape, lower_desc.dtype)
    output_node = ctx.state.add_access(output)

    lower_subset = []
    lower_output_subset = []
    upper_subset = []
    upper_output_subset = []
    for dim_index, size in enumerate(output_desc.shape):
        if dim_index == concat_dim_index:
            lower_subset.append(
                (
                    lower_range_0 - lower.origin[dim_index],
                    lower_range_1 - lower.origin[dim_index] - 1,
                    1,
                )
            )
            upper_subset.append(
                (
                    upper_range_0 - upper.origin[dim_index],
                    upper_range_1 - upper.origin[dim_index] - 1,
                    1,
                )
            )
            # we write the data of the lower range into the output array starting
            # from the index zero
            lower_output_subset.append((0, lower_range_size - 1, 1))
            # the upper range should be written next to the lower range, so the
            # destination subset does not start from index zero
            upper_output_subset.append(
                (
                    lower_range_size,
                    lower_range_size + upper_range_size - 1,
                    1,
                )
            )
        else:
            lower_subset.append(
                (
                    output_domain[dim_index].start - lower.origin[dim_index],
                    output_domain[dim_index].start - lower.origin[dim_index] + size - 1,
                    1,
                )
            )
            upper_subset.append(
                (
                    output_domain[dim_index].start - upper.origin[dim_index],
                    output_domain[dim_index].start - upper.origin[dim_index] + size - 1,
                    1,
                )
            )

            lower_output_subset.append((0, size - 1, 1))
            upper_output_subset.append((0, size - 1, 1))

    ctx.state.add_nedge(
        lower.dc_node,
        output_node,
        dace.Memlet(
            data=lower.dc_node.data,
            subset=dace_subsets.Range(lower_subset),
            other_subset=dace_subsets.Range(lower_output_subset),
        ),
    )
    ctx.state.add_nedge(
        upper.dc_node,
        output_node,
        dace.Memlet(
            data=upper.dc_node.data,
            subset=dace_subsets.Range(upper_subset),
            other_subset=dace_subsets.Range(upper_output_subset),
        ),
    )

    return gtir_to_sdfg_types.FieldopData(output_node, lower.gt_type, origin=tuple(output_origin))


def translate_concat_where(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Lowers a `concat_where` expression to a dataflow where two memlets write
    disjoint subsets, for the lower and upper domain, on one data access node.

    Implements the `PrimitiveTranslator` protocol.
    """
    assert cpm.is_call_to(node, "concat_where")
    assert len(node.args) == 3

    # First argument is a domain expression that defines the mask of the true branch:
    # we extract the dimension along which we need to concatenate the field arguments,
    # and determine whether the true branch argument should be on the lower or upper
    # range with respect to the boundary value.
    mask_domain = gtir_domain.extract_domain(node.args[0])
    if len(mask_domain) != 1:
        raise NotImplementedError("Expected `concat_where` along single axis.")

    # we visit the field arguments for the true and false branch
    tb, fb = (sdfg_builder.visit(node.args[i], ctx=ctx) for i in [1, 2])

    return (
        _translate_concat_where_impl(
            ctx,
            sdfg_builder,
            mask_domain,
            node.annex.domain,
            node.args[1].annex.domain,
            node.args[2].annex.domain,
            tb,
            fb,
        )
        if isinstance(node.type, ts.FieldType)
        else gtx_utils.tree_map(
            lambda _node_domain,
            _tb_node_domain,
            _fb_node_domain,
            _tb_field,
            _fb_field,
            _sdfg_builder=sdfg_builder,
            _ctx=ctx,
            _mask_domain=mask_domain: _translate_concat_where_impl(
                _ctx,
                _sdfg_builder,
                _mask_domain,
                _node_domain,
                _tb_node_domain,
                _fb_node_domain,
                _tb_field,
                _fb_field,
            )
        )(node.annex.domain, node.args[1].annex.domain, node.args[2].annex.domain, tb, fb)
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Union

from lark import lark, lexer as lark_lexer, tree as lark_tree, visitors as lark_visitors

from gt4py.next.iterator import ir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.type_system import type_specifications as ts


GRAMMAR = """
    start: fencil_definition
        | function_definition
        | declaration
        | stencil_closure
        | set_at
        | if_stmt
        | program
        | prec0

    SYM: CNAME
    SYM_REF: CNAME
    TYPE_LITERAL: CNAME
    INT_LITERAL: SIGNED_INT
    FLOAT_LITERAL: SIGNED_FLOAT
    OFFSET_LITERAL: ( INT_LITERAL | CNAME ) "ₒ"
    AXIS_LITERAL: CNAME ("ᵥ" | "ₕ")
    _literal: INT_LITERAL | FLOAT_LITERAL | OFFSET_LITERAL | AXIS_LITERAL
    ID_NAME: CNAME

    ?prec0: prec1
        | "λ(" ( SYM "," )* SYM? ")" "→" prec0 -> lam

    ?prec1: prec2
        | "if" prec1 "then" prec1 "else" prec1 -> ifthenelse

    ?prec2: prec3
        | prec2 "∨" prec3 -> bool_or  

    ?prec3: prec4
        | prec3 "∧" prec4 -> bool_and

    ?prec4: prec5
        | prec4 "==" prec5 -> eq
        | prec4 "<" prec5 -> less
        | prec4 ">" prec5 -> greater

    ?prec5: prec6
        | prec5 "+" prec6 -> plus
        | prec5 "-" prec6 -> minus

    ?prec6: prec7
        | prec6 "×" prec7 -> multiplies
        | prec6 "/" prec7 -> divides

    ?prec7: prec8
        | "·" prec7 -> deref
        | "¬" prec7 -> bool_not
        | "↑" prec7 -> lift
        | "⇑" prec7 -> as_fieldop

    ?prec8: prec9
        | prec8 "[" prec0 "]" -> tuple_get
        | prec8 "(" ( prec0 "," )* prec0? ")" -> call
        | "{" ( prec0 "," )* prec0? "}" -> make_tuple
        | "⟪" ( prec0 "," )* prec0? "⟫" -> shift
        | "u⟨" ( prec0 "," )* prec0? "⟩" -> unstructured_domain
        | "c⟨" ( prec0 "," )* prec0? "⟩" -> cartesian_domain

    ?prec9: _literal
        | SYM_REF
        | named_range
        | "(" prec0 ")"

    ?stmt: set_at | if_stmt
    set_at: prec0 "@" prec0 "←" prec1 ";"
    else_branch_seperator: "else"
    if_stmt: "if" "(" prec0 ")" "{" ( stmt )* "}" else_branch_seperator "{" ( stmt )* "}"

    named_range: AXIS_LITERAL ":" "[" prec0 "," prec0 "["
    function_definition: ID_NAME "=" "λ(" ( SYM "," )* SYM? ")" "→" prec0 ";"
    declaration: ID_NAME "=" "temporary(" "domain=" prec0 "," "dtype=" TYPE_LITERAL ")" ";"
    stencil_closure: prec0 "←" "(" prec0 ")" "(" ( SYM_REF ", " )* SYM_REF ")" "@" prec0 ";"
    fencil_definition: ID_NAME "(" ( SYM "," )* SYM ")" "{" ( function_definition )* ( stencil_closure )+ "}"
    program: ID_NAME "(" ( SYM "," )* SYM ")" "{" ( function_definition )* ( declaration )* ( stmt )+ "}"

    %import common (CNAME, SIGNED_FLOAT, SIGNED_INT, WS)
    %ignore WS
"""  # noqa: RUF001 [ambiguous-unicode-character-string]


@lark_visitors.v_args(inline=True)
class ToIrTransformer(lark_visitors.Transformer):
    def SYM(self, value: lark_lexer.Token) -> ir.Sym:
        return ir.Sym(id=value.value)

    def SYM_REF(self, value: lark_lexer.Token) -> Union[ir.SymRef, ir.Literal]:
        if value.value in ("True", "False"):
            return im.literal(value.value, "bool")
        return ir.SymRef(id=value.value)

    def INT_LITERAL(self, value: lark_lexer.Token) -> ir.Literal:
        return im.literal_from_value(int(value.value))

    def FLOAT_LITERAL(self, value: lark_lexer.Token) -> ir.Literal:
        return im.literal(value.value, "float64")

    def TYPE_LITERAL(self, value: lark_lexer.Token) -> ts.TypeSpec:
        if hasattr(ts.ScalarKind, value.upper()):
            return ts.ScalarType(kind=getattr(ts.ScalarKind, value.upper()))
        raise NotImplementedError(f"Type {value} not supported.")

    def OFFSET_LITERAL(self, value: lark_lexer.Token) -> ir.OffsetLiteral:
        v: Union[int, str] = value.value[:-1]
        try:
            v = int(v)
        except ValueError:
            pass
        return ir.OffsetLiteral(value=v)

    def ID_NAME(self, value: lark_lexer.Token) -> str:
        return value.value

    def AXIS_LITERAL(self, value: lark_lexer.Token) -> ir.AxisLiteral:
        name = value.value[:-1]
        kind = ir.DimensionKind.HORIZONTAL if value.value[-1] == "ₕ" else ir.DimensionKind.VERTICAL
        return ir.AxisLiteral(value=name, kind=kind)

    def lam(self, *args: ir.Node) -> ir.Lambda:
        *params, expr = args
        return ir.Lambda(params=params, expr=expr)

    def bool_and(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="and_"), args=[lhs, rhs])

    def bool_or(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="or_"), args=[lhs, rhs])

    def bool_not(self, arg: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="not_"), args=[arg])

    def bool_xor(self, arg: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="xor_"), args=[arg])

    def plus(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="plus"), args=[lhs, rhs])

    def minus(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="minus"), args=[lhs, rhs])

    def multiplies(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="multiplies"), args=[lhs, rhs])

    def divides(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="divides"), args=[lhs, rhs])

    def floordiv(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="floordiv"), args=[lhs, rhs])

    def mod(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="mod"), args=[lhs, rhs])

    def eq(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="eq"), args=[lhs, rhs])

    def greater(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="greater"), args=[lhs, rhs])

    def less(self, lhs: ir.Expr, rhs: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="less"), args=[lhs, rhs])

    def deref(self, arg: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="deref"), args=[arg])

    def lift(self, arg: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="lift"), args=[arg])

    def as_fieldop(self, arg: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="as_fieldop"), args=[arg])

    def astype(self, arg: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="cast_"), args=[arg])

    def shift(self, *offsets: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="shift"), args=list(offsets))

    def tuple_get(self, tup: ir.Expr, idx: ir.Literal) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="tuple_get"), args=[idx, tup])

    def make_tuple(self, *args: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="make_tuple"), args=list(args))

    def named_range(self, name: ir.AxisLiteral, start: ir.Expr, end: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="named_range"), args=[name, start, end])

    def cartesian_domain(self, *ranges: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="cartesian_domain"), args=list(ranges))

    def unstructured_domain(self, *ranges: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="unstructured_domain"), args=list(ranges))

    def ifthenelse(self, condition: ir.Expr, then: ir.Expr, otherwise: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=ir.SymRef(id="if_"), args=[condition, then, otherwise])

    def call(self, fun: ir.Expr, *args: ir.Expr) -> ir.FunCall:
        return ir.FunCall(fun=fun, args=list(args))

    def function_definition(self, *args: ir.Node) -> ir.FunctionDefinition:
        fid, *params, expr = args
        return ir.FunctionDefinition(id=fid, params=params, expr=expr)

    def if_stmt(self, cond: ir.Expr, *args):
        found_else_seperator = False
        true_branch = []
        false_branch = []
        for arg in args:
            if isinstance(arg, lark_tree.Tree):
                assert arg.data == "else_branch_seperator"
                found_else_seperator = True
                continue

            if not found_else_seperator:
                true_branch.append(arg)
            else:
                false_branch.append(arg)

        return ir.IfStmt(
            cond=cond,
            true_branch=true_branch,
            false_branch=false_branch,
        )

    def declaration(self, *args: ir.Expr) -> ir.Temporary:
        tid, domain, dtype = args
        return ir.Temporary(id=tid, domain=domain, dtype=dtype)

    def set_at(self, *args: ir.Expr) -> ir.SetAt:
        target, domain, expr = args
        return ir.SetAt(expr=expr, domain=domain, target=target)

    def program(self, fid: str, *args: ir.Node) -> ir.Program:
        params = []
        function_definitions = []
        body = []
        declarations = []
        for arg in args:
            if isinstance(arg, ir.Sym):
                params.append(arg)
            elif isinstance(arg, ir.FunctionDefinition):
                function_definitions.append(arg)
            elif isinstance(arg, ir.Temporary):
                declarations.append(arg)
            else:
                assert isinstance(arg, ir.Stmt)
                body.append(arg)
        return ir.Program(
            id=fid,
            function_definitions=function_definitions,
            params=params,
            body=body,
            declarations=declarations,
        )

    def start(self, arg: ir.Node) -> ir.Node:
        return arg


def pparse(pretty_str: str) -> ir.Node:
    parser = lark.Lark(GRAMMAR, parser="earley")
    tree = parser.parse(pretty_str)
    return ToIrTransformer(visit_tokens=True).transform(tree)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import ir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im


# TODO(tehrengruber): This pass can be generalized to all builtins, e.g.
#  `plus((λ(...) → multiplies(...))(...), ...)` can be transformed into
#  `(λ(...) → plus(multiplies(...), ...))(...)`.


class PropagateDeref(PreserveLocationVisitor, NodeTranslator):
    @classmethod
    def apply(cls, node: ir.Node):
        """
        Propagate calls to deref into lambda functions.

        Transform::

            ·(λ(inner_it) → lift(stencil)(inner_it))(outer_it)

        into::

            (λ(inner_it) → ·(lift(stencil)(inner_it)))(outer_it)

        After this pass calls to `deref` are grouped together (better) with lambda functions
        being propagated outwards. This increases the ability of the lift inliner to remove the
        deref + lift combination.
        """
        return cls().visit(node)

    def visit_FunCall(self, node: ir.FunCall):
        if cpm.is_call_to(node, "deref") and cpm.is_let(node.args[0]):
            fun: ir.Lambda = node.args[0].fun
            args: list[ir.Expr] = node.args[0].args
            node = im.let(*zip(fun.params, args, strict=True))(im.deref(fun.expr))
        elif cpm.is_call_to(node, "deref") and cpm.is_call_to(node.args[0], "if_"):
            cond, true_branch, false_branch = node.args[0].args
            return im.if_(cond, im.deref(true_branch), im.deref(false_branch))

        return self.generic_visit(node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""A pretty printer for the functional IR.

Inspired by P. Yelland, “A New Approach to Optimal Code Formatting”, 2015
"""

# TODO(tehrengruber): add support for printing the types of itir.Sym, itir.Literal nodes
from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import Final

from gt4py.eve import NodeTranslator
from gt4py.next.iterator import ir


# replacements for builtin binary operations
BINARY_OPS: Final = {
    "plus": "+",
    "minus": "-",
    "multiplies": "×",  # noqa: RUF001 [ambiguous-unicode-character-string]
    "divides": "/",
    "eq": "==",
    "less": "<",
    "less_equal": "<=",
    "greater": ">",
    "greater_equal": ">=",
    "and_": "∧",
    "or_": "∨",  # noqa: RUF001 [ambiguous-unicode-character-string]
}

# replacements for builtin unary operations
UNARY_OPS: Final = {"deref": "·", "lift": "↑", "not_": "¬", "as_fieldop": "⇑"}

# operator precedence
PRECEDENCE: Final = {
    "__lambda__": 0,
    "if_": 1,
    "or_": 2,
    "and_": 3,
    "eq": 4,
    "less": 4,
    "less_equal": 4,
    "greater": 4,
    "greater_equal": 4,
    "plus": 5,
    "minus": 5,
    "multiplies": 6,
    "divides": 6,
    "deref": 7,
    "not_": 7,
    "lift": 7,
    "as_fieldop": 7,
    "tuple_get": 8,
    "__call__": 8,
}


DEFAULT_INDENT: Final = 2
DEFAULT_WIDTH: Final = 100


class PrettyPrinter(NodeTranslator):
    def __init__(self, indent: int = DEFAULT_INDENT, width: int = DEFAULT_WIDTH) -> None:
        super().__init__()
        self.indent: int = indent
        self.width: int = width

    @staticmethod
    def _hmerge(*blocks: list[str]) -> list[str]:
        if not blocks:
            return []

        def impl(a: list[str], b: list[str]) -> list[str]:
            if not a:
                return b
            if not b:
                return a
            spaces = len(a[-1]) * " "
            return a[:-1] + [a[-1] + b[0]] + [spaces + line for line in b[1:]]

        res = blocks[0]
        for b in blocks[1:]:
            res = impl(res, b)
        return res

    @staticmethod
    def _vmerge(*blocks: list[str]) -> list[str]:
        return [s for b in blocks for s in b]

    def _prec_parens(self, block: list[str], prec: int, op_prec: int) -> list[str]:
        if prec > op_prec:
            return self._hmerge(["("], block, [")"])
        return block

    def _indent(self, block: list[str]) -> list[str]:
        return [" " * self.indent + line for line in block]

    def _cost(self, block: list[str]) -> int:
        if not block:
            return 0
        max_line_length = max(len(line) for line in block)
        return (
            # preferring blocks of fewer lines:
            len(block)
            # strongly preferring blocks with lines not longer than self.width:
            + max(max_line_length - self.width, 0) * 100
            # preferring blocks with lines of uniform length:
            + sum(max_line_length - len(line) for line in block)
        )

    def _optimum(self, a: list[str], b: list[str]) -> list[str]:
        return a if self._cost(a) < self._cost(b) else b

    def _hinterleave(
        self, blocks: Sequence[list[str]], sep: str, *, indent: bool = False
    ) -> Iterator[list[str]]:
        if blocks:
            do_indent = self._indent if indent else lambda x: x
            yield from (do_indent(self._hmerge(block, [sep])) for block in blocks[:-1])
            yield do_indent(blocks[-1])

    def visit_Sym(self, node: ir.Sym, *, prec: int) -> list[str]:
        return [node.id]

    def visit_Literal(self, node: ir.Literal, *, prec: int) -> list[str]:
        return [str(node.value)]

    def visit_InfinityLiteral(self, node: ir.InfinityLiteral, *, prec: int) -> list[str]:
        if node == ir.InfinityLiteral.POSITIVE:
            return ["∞"]
        elif node == ir.InfinityLiteral.NEGATIVE:
            return ["-∞"]
        raise AssertionError()

    def visit_OffsetLiteral(self, node: ir.OffsetLiteral, *, prec: int) -> list[str]:
        return [str(node.value) + "ₒ"]

    def visit_AxisLiteral(self, node: ir.AxisLiteral, *, prec: int) -> list[str]:
        kind = ""
        if node.kind == ir.DimensionKind.HORIZONTAL:
            kind = "ₕ"
        elif node.kind == ir.DimensionKind.VERTICAL:
            kind = "ᵥ"
        return [str(node.value) + kind]

    def visit_SymRef(self, node: ir.SymRef, *, prec: int) -> list[str]:
        return [node.id]

    def visit_Lambda(self, node: ir.Lambda, *, prec: int) -> list[str]:
        params = self.visit(node.params, prec=0)
        expr = self.visit(node.expr, prec=0)

        start, bridge = ["λ("], [") → "]
        if not params:
            params = self._hmerge(start, bridge)
        else:
            hparams = self._hmerge(start, *self._hinterleave(params, ", "), bridge)
            vparams = self._vmerge(start, *self._hinterleave(params, ",", indent=True), bridge)
            params = self._optimum(hparams, vparams)

        hbody = self._hmerge(params, expr)
        vbody = self._vmerge(params, self._indent(expr))
        return self._prec_parens(self._optimum(hbody, vbody), prec, PRECEDENCE["__lambda__"])

    def visit_FunCall(self, node: ir.FunCall, *, prec: int) -> list[str]:
        if isinstance(node.fun, ir.SymRef):
            fun_name = node.fun.id
            if fun_name in BINARY_OPS and len(node.args) == 2:
                # replacing binary ops: plus(x, y) → x + y etc.
                op = BINARY_OPS[fun_name]
                lhs = self.visit(node.args[0], prec=PRECEDENCE[fun_name])
                rhs = self.visit(node.args[1], prec=PRECEDENCE[fun_name] + 1)
                h = self._hmerge(lhs, [" " + op + " "], rhs)
                v = self._vmerge(lhs, self._hmerge([op + " "], rhs))
                return self._prec_parens(self._optimum(h, v), prec, PRECEDENCE[fun_name])
            if fun_name in UNARY_OPS and len(node.args) == 1:
                # replacing unary ops: deref(x) → ·x etc.
                op = UNARY_OPS[fun_name]
                res = self._hmerge([op], self.visit(node.args[0], prec=PRECEDENCE[fun_name]))
                return self._prec_parens(res, prec, PRECEDENCE[fun_name])
            if fun_name == "tuple_get" and len(node.args) == 2:
                # tuple_get(i, x) → x[i]
                idx, tup = self.visit(node.args, prec=PRECEDENCE[fun_name])
                res = self._hmerge(tup, ["["], idx, ["]"])
                return self._prec_parens(res, prec, PRECEDENCE[fun_name])
            if fun_name == "shift":
                shifts = self.visit(node.args, prec=0)
                return self._hmerge(["⟪"], *self._hinterleave(shifts, ", "), ["⟫"])
            if fun_name == "named_range" and len(node.args) == 3:
                # named_range(dim, start, stop) → dim: [star, stop)
                dim, start, end = self.visit(node.args, prec=0)
                res = self._hmerge(
                    dim, [": ["], start, [", "], end, ["["]
                )  # to get matching parenthesis of functions
                return self._prec_parens(res, prec, PRECEDENCE["__call__"])
            if fun_name == "cartesian_domain" and len(node.args) >= 1:
                # cartesian_domain(x, y, ...) → c{ x × y × ... } # noqa: RUF003 [ambiguous-unicode-character-comment]
                args = self.visit(node.args, prec=PRECEDENCE["__call__"])
                return self._hmerge(["c⟨ "], *self._hinterleave(args, ", "), [" ⟩"])
            if fun_name == "unstructured_domain" and len(node.args) >= 1:
                # unstructured_domain(x, y, ...) → u{ x × y × ... } # noqa: RUF003 [ambiguous-unicode-character-comment]
                args = self.visit(node.args, prec=PRECEDENCE["__call__"])
                return self._hmerge(["u⟨ "], *self._hinterleave(args, ", "), [" ⟩"])
            if fun_name == "if_" and len(node.args) == 3:
                # if_(x, y, z) → if x then y else z
                ifb, thenb, elseb = self.visit(node.args, prec=PRECEDENCE["if_"])
                hblocks = self._hmerge(["if "], ifb, [" then "], thenb, [" else "], elseb)
                vblocks = self._vmerge(
                    self._hmerge(["if   "], ifb),
                    self._hmerge(["then "], thenb),
                    self._hmerge(["else "], elseb),
                )
                return self._prec_parens(self._optimum(hblocks, vblocks), prec, PRECEDENCE["if_"])

        fun = self.visit(node.fun, prec=PRECEDENCE["__call__"])
        args = self.visit(node.args, prec=0)

        if not args:
            args = [""]
        else:
            hargs = self._hmerge(*self._hinterleave(args, ", "))
            vargs = self._vmerge(*self._hinterleave(args, ","))
            args = self._optimum(hargs, vargs)

        if node.fun == ir.SymRef(id="make_tuple"):
            # make_tuple(args...) → {args...}
            htup = self._hmerge(["{"], args, ["}"])
            vtup = self._vmerge(["{"], self._indent(args), ["}"])
            return self._optimum(htup, vtup)

        hfun = self._hmerge(fun, ["("], args, [")"])
        vfun = self._vmerge(self._hmerge(fun, ["("]), self._indent(args), [")"])
        return self._prec_parens(self._optimum(hfun, vfun), prec, PRECEDENCE["__call__"])

    def visit_FunctionDefinition(self, node: ir.FunctionDefinition, prec: int) -> list[str]:
        assert prec == 0
        params = self.visit(node.params, prec=0)
        expr = self._hmerge(self.visit(node.expr, prec=0), [";"])

        start, bridge = [node.id + " = λ("], [") → "]
        if not params:
            params = self._hmerge(start, bridge)
        else:
            hparams = self._hmerge(start, *self._hinterleave(params, ", "), bridge)
            vparams = self._vmerge(start, *self._hinterleave(params, ",", indent=True), bridge)
            params = self._optimum(hparams, vparams)

        hbody = self._hmerge(params, expr)
        vbody = self._vmerge(params, self._indent(expr))
        return self._optimum(hbody, vbody)

    def visit_Temporary(self, node: ir.Temporary, *, prec: int) -> list[str]:
        start, end = [node.id + " = temporary("], [");"]
        args = []
        if node.domain is not None:
            args.append(self._hmerge(["domain="], self.visit(node.domain, prec=0)))
        if node.dtype is not None:
            args.append(self._hmerge(["dtype="], [str(node.dtype)]))
        hargs = self._hmerge(*self._hinterleave(args, ", "))
        vargs = self._vmerge(*self._hinterleave(args, ","))
        oargs = self._optimum(hargs, vargs)
        h = self._hmerge(start, oargs, end)
        v = self._vmerge(start, self._indent(oargs), end)
        return self._optimum(h, v)

    def visit_SetAt(self, node: ir.SetAt, *, prec: int) -> list[str]:
        expr = self.visit(node.expr, prec=0)
        domain = self.visit(node.domain, prec=0)
        target = self.visit(node.target, prec=0)

        head = self._hmerge(target, [" @ "], domain)
        foot = self._hmerge([" ← "], expr, [";"])

        h = self._hmerge(head, foot)
        v = self._vmerge(
            head,
            self._indent(self._indent(foot)),
        )
        return self._optimum(h, v)

    def visit_IfStmt(self, node: ir.IfStmt, *, prec: int) -> list[str]:
        cond = self.visit(node.cond, prec=0)
        true_branch = self._vmerge(*self.visit(node.true_branch, prec=0))
        false_branch = self._vmerge(*self.visit(node.false_branch, prec=0))

        hhead = self._hmerge(["if ("], cond, [") {"])
        vhead = self._vmerge(["if ("], cond, [") {"])
        head = self._optimum(hhead, vhead)

        return self._vmerge(
            head, self._indent(true_branch), ["} else {"], self._indent(false_branch), ["}"]
        )

    def visit_Program(self, node: ir.Program, *, prec: int) -> list[str]:
        assert prec == 0
        function_definitions = self.visit(node.function_definitions, prec=0)
        body = self.visit(node.body, prec=0)
        declarations = self.visit(node.declarations, prec=0)
        params = self.visit(node.params, prec=0)

        hparams = self._hmerge([node.id + "("], *self._hinterleave(params, ", "), [") {"])
        vparams = self._vmerge(
            [node.id + "("], *self._hinterleave(params, ",", indent=True), [") {"]
        )
        params = self._optimum(hparams, vparams)

        function_definitions = self._vmerge(*function_definitions)
        declarations = self._vmerge(*declarations)
        body = self._vmerge(*body)

        return self._vmerge(
            params,
            self._indent(function_definitions),
            self._indent(declarations),
            self._indent(body),
            ["}"],
        )

    @classmethod
    def apply(cls, node: ir.Node, indent: int, width: int) -> str:
        return "\n".join(cls(indent=indent, width=width).visit(node, prec=0))


def pformat(x: ir.Node, indent: int = DEFAULT_INDENT, width: int = DEFAULT_WIDTH) -> str:
    return PrettyPrinter.apply(x, indent, width)


def pprint(x: ir.Node, indent: int = DEFAULT_INDENT, width: int = DEFAULT_WIDTH) -> None:
    print(pformat(x, indent, width))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from functools import reduce
from typing import Iterator, Sequence, cast

import gt4py.next.ffront.type_specifications as ts_ffront
import gt4py.next.type_system.type_specifications as ts
from gt4py.next import common
from gt4py.next.type_system import type_info


def _is_zero_dim_field(field: ts.TypeSpec) -> bool:
    return isinstance(field, ts.FieldType) and len(field.dims) == 0


def promote_scalars_to_zero_dim_field(type_: ts.TypeSpec) -> ts.TypeSpec:
    """
    Promote scalar primitive constituents to zero dimensional fields.

    E.g. all elements of a tuple which are scalars are promoted to a zero dimensional field.
    """

    def promote_el(type_el: ts.TypeSpec) -> ts.TypeSpec:
        if isinstance(type_el, ts.ScalarType):
            return ts.FieldType(dims=[], dtype=type_el)
        return type_el

    return type_info.apply_to_primitive_constituents(promote_el, type_)


def promote_zero_dims(
    function_type: ts.FunctionType, args: Sequence[ts.TypeSpec], kwargs: dict[str, ts.TypeSpec]
) -> tuple[tuple, dict]:
    """
    Promote arg types to zero dimensional fields if compatible and required by function signature.

    This function expects the arguments to already be canonicalized using `canonicalize_arguments`.
    """

    def promote_arg(param: ts.TypeSpec, arg: ts.TypeSpec) -> ts.TypeSpec:
        def _as_field(arg_el: ts.TypeSpec, path: tuple[int, ...]) -> ts.TypeSpec:
            param_el = param
            for idx in path:
                if not isinstance(param_el, ts.TupleType):
                    # The parameter has a different structure than the actual argument. Just return
                    # the argument unpromoted and let the further error handling take care of printing
                    # a meaningful error.
                    return arg_el
                param_el = param_el.types[idx]

            if _is_zero_dim_field(param_el) and (
                type_info.is_number(arg_el) or type_info.is_logical(arg_el)
            ):
                if type_info.extract_dtype(param_el) == type_info.extract_dtype(arg_el):
                    return param_el
                else:
                    raise ValueError(f"'{arg_el}' is not compatible with '{param_el}'.")
            return arg_el

        return type_info.apply_to_primitive_constituents(_as_field, arg, with_path_arg=True)

    new_args = [*args]
    for i, (param, arg) in enumerate(
        zip(
            list(function_type.pos_only_args) + list(function_type.pos_or_kw_args.values()),
            args,
            strict=True,
        )
    ):
        new_args[i] = promote_arg(param, arg)
    new_kwargs = {**kwargs}
    for name in set(function_type.kw_only_args.keys()) & set(kwargs.keys()):
        new_kwargs[name] = promote_arg(function_type.kw_only_args[name], kwargs[name])

    return tuple(new_args), new_kwargs


@type_info.return_type.register
def return_type_fieldop(
    fieldop_type: ts_ffront.FieldOperatorType,
    *,
    with_args: list[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
) -> ts.TypeSpec:
    ret_type = type_info.return_type(
        fieldop_type.definition, with_args=with_args, with_kwargs=with_kwargs
    )
    return ret_type


@type_info.canonicalize_arguments.register(ts_ffront.FieldOperatorType)
@type_info.canonicalize_arguments.register(ts_ffront.ProgramType)
def canonicalize_program_or_fieldop_arguments(
    program_type: ts_ffront.ProgramType,
    args: tuple | list,
    kwargs: dict,
    *,
    ignore_errors: bool = False,
    use_signature_ordering: bool = False,
) -> tuple[tuple, dict]:
    return type_info.canonicalize_arguments(
        program_type.definition,
        args,
        kwargs,
        ignore_errors=ignore_errors,
        use_signature_ordering=use_signature_ordering,
    )


@type_info.canonicalize_arguments.register
def canonicalize_scanop_arguments(
    scanop_type: ts_ffront.ScanOperatorType,
    args: tuple | list,
    kwargs: dict,
    *,
    ignore_errors: bool = False,
    use_signature_ordering: bool = False,
) -> tuple[tuple, dict]:
    (_, *cargs), ckwargs = type_info.canonicalize_arguments(
        scanop_type.definition,
        (None, *args),
        kwargs,
        ignore_errors=ignore_errors,
        use_signature_ordering=use_signature_ordering,
    )
    return tuple(cargs), ckwargs


@type_info.function_signature_incompatibilities.register
def function_signature_incompatibilities_fieldop(
    fieldop_type: ts_ffront.FieldOperatorType,
    args: Sequence[ts.TypeSpec],
    kwargs: dict[str, ts.TypeSpec],
) -> Iterator[str]:
    args, kwargs = type_info.canonicalize_arguments(
        fieldop_type.definition, args, kwargs, ignore_errors=True
    )

    error_list = list(
        type_info.structural_function_signature_incompatibilities(
            fieldop_type.definition, args, kwargs
        )
    )
    if len(error_list) > 0:
        yield from error_list
        return

    new_args, new_kwargs = promote_zero_dims(fieldop_type.definition, args, kwargs)
    yield from type_info.function_signature_incompatibilities_func(
        fieldop_type.definition, new_args, new_kwargs, skip_canonicalization=True
    )


def _scan_param_promotion(param: ts.TypeSpec, arg: ts.TypeSpec) -> ts.FieldType | ts.TupleType:
    """
    Promote parameter of a scan pass to match dimensions of respective scan operator argument.

    More specifically: Given a scalar type `param` and a field type `arg` return a field with the
    dtype of the `param` and the dimensions of `arg`. If `param` is a composite of scalars
    the promotion is element-wise.

    Example:
    --------
    >>> _scan_param_promotion(
    ...     ts.ScalarType(kind=ts.ScalarKind.INT64),
    ...     ts.FieldType(
    ...         dims=[common.Dimension("I")], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT64)
    ...     ),
    ... )
    FieldType(dims=[Dimension(value='I', kind=<DimensionKind.HORIZONTAL: 'horizontal'>)], dtype=ScalarType(kind=<ScalarKind.INT64: 8>, shape=None))
    """

    def _as_field(dtype: ts.TypeSpec, path: tuple[int, ...]) -> ts.FieldType:
        assert isinstance(dtype, ts.ScalarType)
        try:
            el_type = reduce(
                lambda type_, idx: type_.types[idx],  # type: ignore[attr-defined]
                path,
                arg,
            )
            return ts.FieldType(dims=type_info.extract_dims(el_type), dtype=dtype)
        except (IndexError, AttributeError):
            # The structure of the scan passes argument and the requested
            # argument type differ. As such we can not extract the dimensions
            # and just return a generic field shown in the error later on.
            # TODO: we want some generic field type here, but our type system does not support it yet.
            return ts.FieldType(dims=[common.Dimension("...")], dtype=dtype)

    res = type_info.apply_to_primitive_constituents(_as_field, param, with_path_arg=True)
    assert isinstance(res, (ts.FieldType, ts.TupleType))
    return res


@type_info.function_signature_incompatibilities.register
def function_signature_incompatibilities_scanop(
    scanop_type: ts_ffront.ScanOperatorType, args: list[ts.TypeSpec], kwargs: dict[str, ts.TypeSpec]
) -> Iterator[str]:
    if not all(
        type_info.is_type_or_tuple_of_type(arg, (ts.ScalarType, ts.FieldType)) for arg in args
    ):
        yield "Arguments to scan operator must be fields, scalars or tuples thereof."
        return

    scan_pass_type: ts.FunctionType = scanop_type.definition
    assert len(scan_pass_type.pos_only_args) == 0

    # canonicalize function arguments
    cargs, ckwargs = type_info.canonicalize_arguments(scanop_type, args, kwargs, ignore_errors=True)

    # check for structural errors
    num_pos_args = len(cargs) - cargs.count(type_info.UNDEFINED_ARG)
    if num_pos_args != len(scan_pass_type.pos_or_kw_args) - 1:
        yield f"Scan operator takes {len(scan_pass_type.pos_or_kw_args) - 1} positional arguments, but {num_pos_args} were given."
        return
    error_list = list(
        type_info.structural_function_signature_incompatibilities(
            scan_pass_type, [None, *cargs], ckwargs
        )
    )
    if len(error_list) > 0:
        yield from error_list
        return
    assert ckwargs.keys() == scan_pass_type.kw_only_args.keys()

    # ensure the dimensions of all arguments can be promoted to a common list of dimensions
    arg_dims = [
        type_info.extract_dims(el)
        for arg in [*cargs, *ckwargs.values()]
        for el in type_info.primitive_constituents(arg)
    ]
    try:
        common.promote_dims(*arg_dims)
    except ValueError as e:
        yield e.args[0]

    assert len(scan_pass_type.pos_only_args) == 0

    # promote parameters
    promoted_params = {}
    for (name, param), arg in zip(list(scan_pass_type.pos_or_kw_args.items())[1:], cargs):
        promoted_params[name] = _scan_param_promotion(param, arg)
    promoted_kwparams = {}
    for name, param, arg in zip(
        ckwargs.keys(), scan_pass_type.kw_only_args.values(), ckwargs.values()
    ):
        promoted_kwparams[name] = _scan_param_promotion(param, arg)

    # build a function type to leverage the already existing signature checking capabilities
    function_type = ts.FunctionType(
        pos_only_args=[],
        pos_or_kw_args=promoted_params,
        kw_only_args=promoted_kwparams,
        returns=ts.DeferredType(constraint=None),
    )

    yield from type_info.function_signature_incompatibilities_func(
        function_type,
        *promote_zero_dims(function_type, cargs, ckwargs),
        skip_canonicalization=True,
        skip_structural_checks=True,
    )


@type_info.function_signature_incompatibilities.register
def function_signature_incompatibilities_program(
    program_type: ts_ffront.ProgramType, args: tuple[ts.TypeSpec], kwargs: dict[str, ts.TypeSpec]
) -> Iterator[str]:
    args, kwargs = type_info.canonicalize_arguments(
        program_type.definition, args, kwargs, ignore_errors=True
    )

    error_list = list(
        type_info.structural_function_signature_incompatibilities(
            program_type.definition, args, kwargs
        )
    )
    if len(error_list) > 0:
        yield from error_list
        return

    new_args, new_kwargs = promote_zero_dims(program_type.definition, args, kwargs)
    yield from type_info.function_signature_incompatibilities_func(
        program_type.definition, new_args, new_kwargs, skip_canonicalization=True
    )


@type_info.return_type.register
def return_type_scanop(
    callable_type: ts_ffront.ScanOperatorType,
    *,
    with_args: list[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
) -> ts.TypeSpec:
    carry_dtype = callable_type.definition.returns
    promoted_dims = common.promote_dims(
        *(
            type_info.extract_dims(el)
            for arg in with_args + list(with_kwargs.values())
            for el in type_info.primitive_constituents(arg)
        ),
        # the vertical axis is always added to the dimension of the returned
        #  field
        [callable_type.axis],
    )
    return type_info.apply_to_primitive_constituents(
        lambda arg: ts.FieldType(dims=promoted_dims, dtype=cast(ts.ScalarType, arg)), carry_dtype
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

from gt4py.eve import utils as eve_utils
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.transforms.collapse_tuple import CollapseTuple
from gt4py.next.iterator.transforms.constant_folding import ConstantFolding
from gt4py.next.iterator.transforms.inline_lambdas import InlineLambdas


def dead_code_elimination(
    program: itir.Program,
    *,
    collapse_tuple_uids: Optional[eve_utils.UIDGenerator] = None,
    offset_provider_type: common.OffsetProviderType,
) -> itir.Program:
    """
    Perform dead code elimination on a program by simplifying or removing
    unused or unreachable code constructs.

    This transformation is a composition of InlineLambdas, ConstantFolding, and CollapseTuple.

    Example:
    ```
    let val = {inp1, inp2}[0]
      if_ True then
        val
      else
        inp3
    end
    ```
    is transformed into
    ```
    inp1
    ```
    """
    # ensure all constant let bindings are inlined
    # `let var=True in if_(var, val1, val2) end` -> `if_(True, val1, val2)`
    # TODO(tehrengruber): we only want to inline literals here
    program = InlineLambdas.apply(program, opcount_preserving=True)

    # remove the unreachable if branches
    # e.g. `if_(True, val1, val2)` -> `val1`
    program = ConstantFolding.apply(program)  # type: ignore[assignment]  # always an itir.Program

    # inline again since after constant folding some expressions might not be referenced anymore,
    # e.g. `let field = as_fieldop(...) in val1 end` -> `val1`.
    # TODO(tehrengruber): If we first re-arrange the tree such that let bindings are placed as
    #  close as possible to references to them we don't need to inline again here.
    # note: `force_inline_lambda_args` increases the size of the tree and may not be required for
    # dead-code-elimination, but is needed later since the domain inference cannot handle "user"
    # functions, e.g. `let f = λ(...) → ... in f(...)`
    program = InlineLambdas.apply(program, opcount_preserving=True, force_inline_lambda_args=True)

    # get rid of tuple elements that are never accessed
    # `{a, b}[0]` -> `a`
    program = CollapseTuple.apply(
        program,
        enabled_transformations=~CollapseTuple.Transformation.PROPAGATE_TO_IF_ON_TUPLES,
        uids=collapse_tuple_uids,
        offset_provider_type=offset_provider_type,
    )  # type: ignore[assignment]  # always an itir.Program

    return program
